"""
    :codeauthor: :email:`Anthony Shaw <anthonyshaw@apache.org>`
"""


import salt.modules.napalm_network as napalm_network
import salt.modules.napalm_yang_mod as napalm_yang_mod
import tests.support.napalm as napalm_test_support
from salt.utils.immutabletypes import freeze
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MockNapalmYangModel:
    def Root(self):
        return MagicMock()


class MockNapalmYangModels:
    openconfig_interfaces = MockNapalmYangModel()


class MockUtils:
    def __init__(self, test_diff):
        self.test_diff = test_diff

    def diff(self, *args):
        return self.test_diff


class MockNapalmYangModule:
    def __init__(self, test_diff):
        self.base = MockNapalmYangModel()
        self.models = MockNapalmYangModels()
        self.utils = MockUtils(test_diff)


class NapalmYangModModuleTestCase(TestCase, LoaderModuleMockMixin):
    @classmethod
    def setUpClass(cls):
        cls._test_config = freeze(
            {
                "comment": "Configuration discarded.",
                "already_configured": False,
                "result": True,
                "diff": (
                    '[edit interfaces xe-0/0/5]+   description "Adding a description";'
                ),
            }
        )
        cls._test_diff = freeze({"diff1": "value"})

    @classmethod
    def tearDownClass(cls):
        cls._test_config = cls._test_diff = None

    def setup_loader_modules(self):
        patcher = patch(
            "salt.utils.napalm.get_device",
            MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        def mock_net_load_config(**kwargs):
            return self._test_config.copy()

        module_globals = {
            "__salt__": {
                "config.get": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                ),
                "file.file_exists": napalm_test_support.true,
                "file.join": napalm_test_support.join,
                "file.get_managed": napalm_test_support.get_managed_file,
                "random.hash": napalm_test_support.random_hash,
                "net.load_template": napalm_network.load_template,
                "net.load_config": mock_net_load_config,
            },
            "napalm_yang": MockNapalmYangModule(self._test_diff.copy()),
        }

        return {napalm_yang_mod: module_globals, napalm_network: module_globals}

    def test_diff(self):
        ret = napalm_yang_mod.diff({}, {"test": True}, "models.openconfig_interfaces")
        assert ret == self._test_diff.copy()

    def test_diff_list(self):
        """
        Test it with an actual list
        """
        ret = napalm_yang_mod.diff({}, {"test": True}, ["models.openconfig_interfaces"])
        assert ret == self._test_diff.copy()

    def test_parse(self):
        ret = napalm_yang_mod.parse("models.openconfig_interfaces")
        assert ret is not None

    def test_get_config(self):
        ret = napalm_yang_mod.get_config({}, "models.openconfig_interfaces")
        assert ret is not None

    def test_load_config(self):
        ret = napalm_yang_mod.load_config({}, "models.openconfig_interfaces")
        assert ret == self._test_config.copy()

    def test_compliance_report(self):
        ret = napalm_yang_mod.compliance_report({}, "models.openconfig_interfaces")
        assert ret is not None



"""
This is the default ipcidr matcher.
"""

import logging

import salt.utils.network
from salt._compat import ipaddress

log = logging.getLogger(__name__)


def match(tgt, opts=None, minion_id=None):
    """
    Matches based on IP address or CIDR notation
    """
    if not opts:
        opts = __opts__

    try:
        # Target is an address?
        tgt = ipaddress.ip_address(tgt)
    except:  # pylint: disable=bare-except
        try:
            # Target is a network?
            tgt = ipaddress.ip_network(tgt)
        except:  # pylint: disable=bare-except
            log.error("Invalid IP/CIDR target: %s", tgt)
            return []
    proto = "ipv{}".format(tgt.version)

    grains = opts["grains"]

    if proto not in grains:
        match = False
    elif isinstance(tgt, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
        match = str(tgt) in grains[proto]
    else:
        match = salt.utils.network.in_subnet(tgt, grains[proto])

    return match



"""
Manage RabbitMQ Virtual Hosts
=============================

Example:

.. code-block:: yaml

    virtual_host:
      rabbitmq_vhost.present:
        - user: rabbit_user
        - conf: .*
        - write: .*
        - read: .*
"""


import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if RabbitMQ is installed.
    """
    if salt.utils.path.which("rabbitmqctl"):
        return True
    return (False, "Command not found: rabbitmqctl")


def present(name):
    """
    Ensure the RabbitMQ VHost exists.

    name
        VHost name

    user
        Initial user permission to set on the VHost, if present

        .. deprecated:: 2015.8.0
    owner
        Initial owner permission to set on the VHost, if present

        .. deprecated:: 2015.8.0
    conf
        Initial conf string to apply to the VHost and user. Defaults to .*

        .. deprecated:: 2015.8.0
    write
        Initial write permissions to apply to the VHost and user.
        Defaults to .*

        .. deprecated:: 2015.8.0
    read
        Initial read permissions to apply to the VHost and user.
        Defaults to .*

        .. deprecated:: 2015.8.0
    runas
        Name of the user to run the command

        .. deprecated:: 2015.8.0
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    vhost_exists = __salt__["rabbitmq.vhost_exists"](name)

    if vhost_exists:
        ret["comment"] = "Virtual Host '{}' already exists.".format(name)
        return ret

    if not __opts__["test"]:
        result = __salt__["rabbitmq.add_vhost"](name)
        if "Error" in result:
            ret["result"] = False
            ret["comment"] = result["Error"]
            return ret
        elif "Added" in result:
            ret["comment"] = result["Added"]

    # If we've reached this far before returning, we have changes.
    ret["changes"] = {"old": "", "new": name}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Virtual Host '{}' will be created.".format(name)

    return ret


def absent(name):
    """
    Ensure the RabbitMQ Virtual Host is absent

    name
        Name of the Virtual Host to remove
    runas
        User to run the command

        .. deprecated:: 2015.8.0
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    vhost_exists = __salt__["rabbitmq.vhost_exists"](name)

    if not vhost_exists:
        ret["comment"] = "Virtual Host '{}' is not present.".format(name)
        return ret

    if not __opts__["test"]:
        result = __salt__["rabbitmq.delete_vhost"](name)
        if "Error" in result:
            ret["result"] = False
            ret["comment"] = result["Error"]
            return ret
        elif "Deleted" in result:
            ret["comment"] = result["Deleted"]

    # If we've reached this far before returning, we have changes.
    ret["changes"] = {"new": "", "old": name}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Virtual Host '{}' will be removed.".format(name)

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest
import salt.states.boto_route53 as boto_route53
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_route53: {}}


def test_present():
    """
    Test to ensure the Route53 record is present.
    """
    name = "test.example.com."
    value = "1.1.1.1"
    zone = "example.com."
    record_type = "A"

    ret = {"name": name, "result": False, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[{}, {}, {"value": ""}, False])
    mock_bool = MagicMock(return_value=False)
    with patch.dict(
        boto_route53.__salt__,
        {"boto_route53.get_record": mock, "boto_route53.add_record": mock_bool},
    ):
        with patch.dict(boto_route53.__opts__, {"test": False}):
            comt = "Failed to add {} Route53 record.".format(name)
            ret.update({"comment": comt})
            assert boto_route53.present(name, value, zone, record_type) == ret

        with patch.dict(boto_route53.__opts__, {"test": True}):
            comt = "Route53 record {} set to be added.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_route53.present(name, value, zone, record_type) == ret

            comt = "Route53 record {} set to be updated.".format(name)
            ret.update({"comment": comt})
            assert boto_route53.present(name, value, zone, record_type) == ret

        ret.update({"comment": "", "result": True})
        assert boto_route53.present(name, value, zone, record_type) == ret


def test_absent():
    """
    Test to ensure the Route53 record is deleted.
    """
    name = "test.example.com."
    zone = "example.com."
    record_type = "A"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(boto_route53.__salt__, {"boto_route53.get_record": mock}):
        comt = "{} does not exist.".format(name)
        ret.update({"comment": comt})
        assert boto_route53.absent(name, zone, record_type) == ret

        with patch.dict(boto_route53.__opts__, {"test": True}):
            comt = "Route53 record {} set to be deleted.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_route53.absent(name, zone, record_type) == ret



import logging
import warnings

import salt.utils.url
from salt.serializers.yamlex import deserialize

log = logging.getLogger(__name__)


def render(sls_data, saltenv="base", sls="", **kws):
    """
    Accepts YAML_EX as a string or as a file object and runs it through the YAML_EX
    parser.

    :rtype: A Python data structure
    """
    with warnings.catch_warnings(record=True) as warn_list:
        data = deserialize(sls_data) or {}

        for item in warn_list:
            log.warning(
                "%s found in %s saltenv=%s",
                item.message,
                salt.utils.url.create(sls),
                saltenv,
            )

        log.debug("Results of SLS rendering: \n%s", data)

    return data



"""
Integration Tests for restcherry salt-api with pam eauth
"""
import urllib.parse

import pytest
import salt.utils.platform
import tests.support.cherrypy_testclasses as cptc
from tests.support.case import ModuleCase
from tests.support.unit import skipIf

if cptc.HAS_CHERRYPY:
    import cherrypy

USERA = "saltdev-netapi"
USERA_PWD = "saltdev"
HASHED_USERA_PWD = "$6$SALTsalt$ZZFD90fKFWq8AGmmX0L3uBtS9fXL62SrTk5zcnQ6EkD6zoiM3kB88G1Zvs0xm/gZ7WXJRs5nsTBybUvGSqZkT."

AUTH_CREDS = {"username": USERA, "password": USERA_PWD, "eauth": "pam"}


@skipIf(cptc.HAS_CHERRYPY is False, "CherryPy not installed")
class TestAuthPAM(cptc.BaseRestCherryPyTest, ModuleCase):
    """
    Test auth with pam using salt-api
    """

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    def setUp(self):
        super().setUp()
        try:
            add_user = self.run_function("user.add", [USERA], createhome=False)
            add_pwd = self.run_function(
                "shadow.set_password",
                [
                    USERA,
                    USERA_PWD if salt.utils.platform.is_darwin() else HASHED_USERA_PWD,
                ],
            )
            self.assertTrue(add_user)
            self.assertTrue(add_pwd)
            user_list = self.run_function("user.list_users")
            self.assertIn(USERA, str(user_list))
        except AssertionError:
            self.run_function("user.delete", [USERA], remove=True)
            self.skipTest("Could not add user or password, skipping test")

    @pytest.mark.slow_test
    def test_bad_pwd_pam_chsh_service(self):
        """
        Test login while specifying chsh service with bad passwd
        This test ensures this PR is working correctly:
        https://github.com/saltstack/salt/pull/31826
        """
        copyauth_creds = AUTH_CREDS.copy()
        copyauth_creds["service"] = "chsh"
        copyauth_creds["password"] = "wrong_password"
        body = urllib.parse.urlencode(copyauth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "401 Unauthorized")

    @pytest.mark.slow_test
    def test_bad_pwd_pam_login_service(self):
        """
        Test login while specifying login service with bad passwd
        This test ensures this PR is working correctly:
        https://github.com/saltstack/salt/pull/31826
        """
        copyauth_creds = AUTH_CREDS.copy()
        copyauth_creds["service"] = "login"
        copyauth_creds["password"] = "wrong_password"
        body = urllib.parse.urlencode(copyauth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "401 Unauthorized")

    @pytest.mark.slow_test
    def test_good_pwd_pam_chsh_service(self):
        """
        Test login while specifying chsh service with good passwd
        This test ensures this PR is working correctly:
        https://github.com/saltstack/salt/pull/31826
        """
        copyauth_creds = AUTH_CREDS.copy()
        copyauth_creds["service"] = "chsh"
        body = urllib.parse.urlencode(copyauth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")

    @pytest.mark.slow_test
    def test_good_pwd_pam_login_service(self):
        """
        Test login while specifying login service with good passwd
        This test ensures this PR is working correctly:
        https://github.com/saltstack/salt/pull/31826
        """
        copyauth_creds = AUTH_CREDS.copy()
        copyauth_creds["service"] = "login"
        body = urllib.parse.urlencode(copyauth_creds)
        request, response = self.request(
            "/login",
            method="POST",
            body=body,
            headers={"content-type": "application/x-www-form-urlencoded"},
        )
        self.assertEqual(response.status, "200 OK")

    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    def tearDown(self):
        """
        Clean up after tests. Delete user
        """
        super().tearDown()
        user_list = self.run_function("user.list_users")
        # Remove saltdev user
        if USERA in user_list:
            self.run_function("user.delete", [USERA], remove=True)
        # need to exit cherypy engine
        cherrypy.engine.exit()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.win_disk as win_disk
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class MockKernel32:
    """
    Mock windll class
    """

    def __init__(self):
        pass

    @staticmethod
    def GetLogicalDrives():
        """
        Mock GetLogicalDrives method
        """
        return 1


class MockWindll:
    """
    Mock windll class
    """

    def __init__(self):
        self.kernel32 = MockKernel32()


class MockCtypes:
    """
    Mock ctypes class
    """

    def __init__(self):
        self.windll = MockWindll()


class WinDiskTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_disk
    """

    def setup_loader_modules(self):
        return {win_disk: {"ctypes": MockCtypes()}}

    # 'usage' function tests: 1

    def test_usage(self):
        """
        Test if it return usage information for volumes mounted on this minion.
        """
        self.assertDictEqual(
            win_disk.usage(),
            {
                "A:\\": {
                    "available": None,
                    "1K-blocks": None,
                    "used": None,
                    "capacity": None,
                    "filesystem": "A:\\",
                }
            },
        )



"""
unit tests for the sqs_events engine
"""

import salt.engines.sqs_events as sqs_events
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf


@skipIf(sqs_events.HAS_BOTO is False, "The boto library is not installed")
class EngineSqsEventTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.engine.sqs_events
    """

    def setup_loader_modules(self):
        patcher = patch("salt.engines.sqs_events.boto.sqs")
        self.mock_sqs = patcher.start()
        self.addCleanup(patcher.stop)
        self.addCleanup(delattr, self, "mock_sqs")
        return {sqs_events: {}}

    def sample_msg(self):
        fake_msg = MagicMock()
        fake_msg.get_body.return_value = "This is a test message"
        fake_msg.delete.return_value = True
        return fake_msg

    # 'present' function tests: 1
    def test_no_queue_present(self):
        """
        Test to ensure the SQS engine logs a warning when queue not present
        """
        with patch("salt.engines.sqs_events.log") as mock_logging:
            with patch("time.sleep", return_value=None) as mock_sleep:
                q = None
                q_name = "mysqs"
                mock_fire = MagicMock(return_value=True)
                sqs_events._process_queue(q, q_name, mock_fire)
                self.assertTrue(mock_logging.warning.called)
                self.assertFalse(self.mock_sqs.queue.Queue().get_messages.called)

    def test_minion_message_fires(self):
        """
        Test SQS engine correctly gets and fires messages on minion
        """
        msgs = [self.sample_msg(), self.sample_msg()]
        self.mock_sqs.queue.Queue().get_messages.return_value = msgs
        q = self.mock_sqs.queue.Queue()
        q_name = "mysqs"
        mock_event = MagicMock(return_value=True)
        mock_fire = MagicMock(return_value=True)
        with patch.dict(sqs_events.__salt__, {"event.send": mock_event}):
            sqs_events._process_queue(q, q_name, mock_fire)
            self.assertTrue(self.mock_sqs.queue.Queue().get_messages.called)
            self.assertTrue(all(x.delete.called for x in msgs))

    def test_master_message_fires(self):
        """
        Test SQS engine correctly gets and fires messages on master
        """
        msgs = [self.sample_msg(), self.sample_msg()]
        self.mock_sqs.queue.Queue().get_messages.return_value = msgs
        q = self.mock_sqs.queue.Queue()
        q_name = "mysqs"
        mock_fire = MagicMock(return_value=True)
        sqs_events._process_queue(q, q_name, mock_fire)
        self.assertTrue(self.mock_sqs.queue.Queue().get_messages.called, len(msgs))
        self.assertTrue(mock_fire.called, len(msgs))



"""
Store key/value pairs in a CSV file

.. versionadded:: 2016.11.0

Example configuration:

.. code-block:: yaml

    ext_pillar:
      - csv: /path/to/file.csv

    # or

    ext_pillar:
      - csv:
          path: /path/to/file.csv
          namespace: 'subkey'
          fieldnames:
          - col1
          - col2
          - col2

The first column must be minion IDs and the first row must be dictionary keys.
E.g.:

==========  =========   ======
id          role        env
==========  =========   ======
jerry       web         prod
stuart      web         stage
dave        web         qa
phil        db          prod
kevin       db          stage
mike        db          qa
==========  =========   ======

Will produce the following Pillar values for a minion named "jerry":

.. code-block:: python

    {
        'role': 'web',
        'env': 'prod',
    }
"""

import csv

import salt.utils.files

__virtualname__ = "csv"


def __virtual__():
    return __virtualname__


def ext_pillar(
    mid,
    pillar,
    path,
    idkey="id",
    namespace=None,
    fieldnames=None,
    restkey=None,
    restval=None,
    dialect="excel",
):
    """
    Read a CSV into Pillar

    :param str path: Absolute path to a CSV file.
    :param str idkey: (Optional) The column name of minion IDs.
    :param str namespace: (Optional) A pillar key to namespace the values under.
    :param list fieldnames: (Optional) if the first row of the CSV is not
        column names they may be specified here instead.
    """
    with salt.utils.files.fopen(path, "r") as f:
        sheet = csv.DictReader(
            f, fieldnames, restkey=restkey, restval=restval, dialect=dialect
        )

        for row in sheet:
            if row[idkey] == mid:
                if namespace:
                    return {namespace: row}
                else:
                    return row

    return {}



"""
salt-ssh testing
"""

import os
import shutil

import pytest
from tests.support.case import SSHCase
from tests.support.runtests import RUNTIME_VARS


class SSHTest(SSHCase):
    """
    Test general salt-ssh functionality
    """

    def setUp(self):
        thin_dir = self.run_function("config.get", ["thin_dir"], wipe=False)
        self.addCleanup(shutil.rmtree, thin_dir, ignore_errors=True)

    @pytest.mark.slow_test
    def test_ping(self):
        """
        Test a simple ping
        """
        ret = self.run_function("test.ping")
        self.assertTrue(ret, "Ping did not return true")

    @pytest.mark.slow_test
    def test_thin_dir(self):
        """
        test to make sure thin_dir is created
        and salt-call file is included
        """
        thin_dir = self.run_function("config.get", ["thin_dir"], wipe=False)
        os.path.isdir(thin_dir)
        os.path.exists(os.path.join(thin_dir, "salt-call"))
        os.path.exists(os.path.join(thin_dir, "running_data"))

    @pytest.mark.slow_test
    def test_set_path(self):
        """
        test setting the path env variable
        """
        path = "/pathdoesnotexist/"
        roster = os.path.join(RUNTIME_VARS.TMP, "roster-set-path")
        self.custom_roster(
            roster, data={"set_path": "$PATH:/usr/local/bin/:{}".format(path)}
        )
        ret = self.run_function("environ.get", ["PATH"], roster_file=roster)
        assert path in ret

    @pytest.mark.slow_test
    def test_tty(self):
        """
        test using tty
        """
        roster = os.path.join(RUNTIME_VARS.TMP, "roster-tty")
        self.custom_roster(roster, data={"tty": True})
        ret = self.run_function("test.ping", roster_file=roster)
        assert ret is True



"""
    :codeauthor: :email:`Jayesh Kariya <jayeshk@saltstack.com>`
"""

import os

import pytest
import salt.modules.beacons as beacons
from salt.utils.event import SaltEvent
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class BeaconsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.beacons
    """

    @classmethod
    def setUpClass(cls):
        cls.sock_dir = os.path.join(RUNTIME_VARS.TMP, "test-socks")

    def setup_loader_modules(self):
        return {beacons: {}}

    @pytest.mark.slow_test
    def test_delete(self):
        """
        Test deleting a beacon.
        """
        comm1 = "Deleted beacon: ps."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_delete_complete",
                "beacons": {},
            },
        ]

        with patch.dict(
            beacons.__opts__,
            {
                "beacons": {
                    "ps": [
                        {"processes": {"salt-master": "stopped", "apache2": "stopped"}}
                    ]
                },
                "sock_dir": self.sock_dir,
            },
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.delete("ps"), {"comment": comm1, "result": True}
                    )

    @pytest.mark.slow_test
    def test_add(self):
        """
        Test adding a beacon
        """
        comm1 = "Added beacon: ps."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_list_complete",
                "beacons": {},
            },
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_list_available_complete",
                "beacons": ["ps"],
            },
            {
                "complete": True,
                "valid": True,
                "vcomment": "",
                "tag": "/salt/minion/minion_beacons_list_complete",
            },
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacon_add_complete",
                "beacons": {
                    "ps": [
                        {"processes": {"salt-master": "stopped", "apache2": "stopped"}}
                    ]
                },
            },
        ]

        with patch.dict(beacons.__opts__, {"beacons": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.add(
                            "ps",
                            [
                                {
                                    "processes": {
                                        "salt-master": "stopped",
                                        "apache2": "stopped",
                                    }
                                }
                            ],
                        ),
                        {"comment": comm1, "result": True},
                    )

    @pytest.mark.slow_test
    def test_save(self):
        """
        Test saving beacons.
        """
        comm1 = "Beacons saved to {}beacons.conf.".format(RUNTIME_VARS.TMP + os.sep)
        with patch.dict(
            beacons.__opts__,
            {
                "conf_file": os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "foo"),
                "beacons": {},
                "default_include": RUNTIME_VARS.TMP + os.sep,
                "sock_dir": self.sock_dir,
            },
        ):

            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "beacons": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    self.assertDictEqual(
                        beacons.save(), {"comment": comm1, "result": True}
                    )

    @pytest.mark.slow_test
    def test_disable(self):
        """
        Test disabling beacons
        """
        comm1 = "Disabled beacons on minion."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_disabled_complete",
                "beacons": {
                    "enabled": False,
                    "ps": [
                        {"processes": {"salt-master": "stopped", "apache2": "stopped"}}
                    ],
                },
            }
        ]

        with patch.dict(beacons.__opts__, {"beacons": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.disable(), {"comment": comm1, "result": True}
                    )

    @pytest.mark.slow_test
    def test_enable(self):
        """
        Test enabling beacons
        """
        comm1 = "Enabled beacons on minion."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacon_enabled_complete",
                "beacons": {
                    "enabled": True,
                    "ps": [
                        {"processes": {"salt-master": "stopped", "apache2": "stopped"}}
                    ],
                },
            }
        ]

        with patch.dict(beacons.__opts__, {"beacons": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.enable(), {"comment": comm1, "result": True}
                    )

    @pytest.mark.slow_test
    def test_add_beacon_module(self):
        """
        Test adding a beacon
        """
        comm1 = "Added beacon: watch_salt_master."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_list_complete",
                "beacons": {},
            },
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_list_available_complete",
                "beacons": ["ps"],
            },
            {
                "complete": True,
                "valid": True,
                "vcomment": "",
                "tag": "/salt/minion/minion_beacons_list_complete",
            },
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacon_add_complete",
                "beacons": {
                    "watch_salt_master": [
                        {"processes": {"salt-master": "stopped"}},
                        {"beacon_module": "ps"},
                    ]
                },
            },
        ]

        with patch.dict(beacons.__opts__, {"beacons": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.add(
                            "watch_salt_master",
                            [
                                {"processes": {"salt-master": "stopped"}},
                                {"beacon_module": "ps"},
                            ],
                        ),
                        {"comment": comm1, "result": True},
                    )

    @pytest.mark.slow_test
    def test_enable_beacon_module(self):
        """
        Test enabling beacons
        """
        comm1 = "Enabled beacons on minion."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacon_enabled_complete",
                "beacons": {
                    "enabled": True,
                    "watch_salt_master": [
                        {"processes": {"salt-master": "stopped"}},
                        {"beacon_module": "ps"},
                    ],
                },
            }
        ]

        with patch.dict(beacons.__opts__, {"beacons": {}, "sock_dir": self.sock_dir}):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.enable(), {"comment": comm1, "result": True}
                    )

    @pytest.mark.slow_test
    def test_delete_beacon_module(self):
        """
        Test deleting a beacon.
        """
        comm1 = "Deleted beacon: watch_salt_master."
        event_returns = [
            {
                "complete": True,
                "tag": "/salt/minion/minion_beacons_delete_complete",
                "beacons": {},
            },
        ]

        with patch.dict(
            beacons.__opts__,
            {
                "beacons": {
                    "watch_salt_master": [
                        {"processes": {"salt-master": "stopped"}},
                        {"beacon_module": "ps"},
                    ]
                },
                "sock_dir": self.sock_dir,
            },
        ):
            mock = MagicMock(return_value=True)
            with patch.dict(beacons.__salt__, {"event.fire": mock}):
                with patch.object(SaltEvent, "get_event", side_effect=event_returns):
                    self.assertDictEqual(
                        beacons.delete("watch_salt_master"),
                        {"comment": comm1, "result": True},
                    )



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""
import copy
import glob
import os

import pytest
import salt.config
import salt.loader
import salt.modules.win_lgpo as win_lgpo
import salt.states.win_lgpo
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase, skipIf


class WinLGPOTestCase(TestCase):
    """
    Test cases for salt.modules.win_lgpo
    """

    encoded_null = chr(0).encode("utf-16-le")

    def test__getAdmlDisplayName(self):
        display_name = "$(string.KeepAliveTime1)"
        adml_xml_data = "junk, we are mocking the return"
        obj_xpath = Mock()
        obj_xpath.text = "300000 or 5 minutes (recommended) "
        mock_xpath_obj = MagicMock(return_value=[obj_xpath])
        with patch.object(win_lgpo, "ADML_DISPLAY_NAME_XPATH", mock_xpath_obj):
            result = win_lgpo._getAdmlDisplayName(
                adml_xml_data=adml_xml_data, display_name=display_name
            )
        expected = "300000 or 5 minutes (recommended)"
        self.assertEqual(result, expected)

    def test__regexSearchKeyValueCombo_enabled(self):
        """
        Make sure
        """
        policy_data = (
            b"[\x00s\x00o\x00f\x00t\x00w\x00a\x00r\x00e\x00\\\x00p"
            b"\x00o\x00l\x00i\x00c\x00i\x00e\x00s\x00\\\x00m\x00i"
            b"\x00c\x00r\x00o\x00s\x00o\x00f\x00t\x00\\\x00w\x00i"
            b"\x00n\x00d\x00o\x00w\x00s\x00\\\x00w\x00i\x00n\x00d"
            b"\x00o\x00w\x00s\x00 \x00e\x00r\x00r\x00o\x00r\x00 "
            b"\x00r\x00e\x00p\x00o\x00r\x00t\x00i\x00n\x00g\x00\\"
            b"\x00c\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00\x00;\x00D"
            b"\x00e\x00f\x00a\x00u\x00l\x00t\x00C\x00o\x00n\x00s"
            b"\x00e\x00n\x00t\x00\x00\x00;\x00\x01\x00\x00\x00;\x00"
            b"\x04\x00\x00\x00;\x00\x02\x00\x00\x00]\x00"
        )
        policy_regpath = (
            b"\x00s\x00o\x00f\x00t\x00w\x00a\x00r\x00e\x00\\\x00p"
            b"\x00o\x00l\x00i\x00c\x00i\x00e\x00s\x00\\\x00m\x00i"
            b"\x00c\x00r\x00o\x00s\x00o\x00f\x00t\x00\\\x00w\x00i"
            b"\x00n\x00d\x00o\x00w\x00s\x00\\\x00w\x00i\x00n\x00d"
            b"\x00o\x00w\x00s\x00 \x00e\x00r\x00r\x00o\x00r\x00 "
            b"\x00r\x00e\x00p\x00o\x00r\x00t\x00i\x00n\x00g\x00\\"
            b"\x00c\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00"
        )
        policy_regkey = (
            b"\x00D\x00e\x00f\x00a\x00u\x00l\x00t\x00C\x00o\x00n"
            b"\x00s\x00e\x00n\x00t\x00\x00"
        )
        test = win_lgpo._regexSearchKeyValueCombo(
            policy_data=policy_data,
            policy_regpath=policy_regpath,
            policy_regkey=policy_regkey,
        )
        self.assertEqual(test, policy_data)

    def test__regexSearchKeyValueCombo_not_configured(self):
        """
        Make sure
        """
        policy_data = b""
        policy_regpath = (
            b"\x00s\x00o\x00f\x00t\x00w\x00a\x00r\x00e\x00\\\x00p"
            b"\x00o\x00l\x00i\x00c\x00i\x00e\x00s\x00\\\x00m\x00i"
            b"\x00c\x00r\x00o\x00s\x00o\x00f\x00t\x00\\\x00w\x00i"
            b"\x00n\x00d\x00o\x00w\x00s\x00\\\x00w\x00i\x00n\x00d"
            b"\x00o\x00w\x00s\x00 \x00e\x00r\x00r\x00o\x00r\x00 "
            b"\x00r\x00e\x00p\x00o\x00r\x00t\x00i\x00n\x00g\x00\\"
            b"\x00c\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00"
        )
        policy_regkey = (
            b"\x00D\x00e\x00f\x00a\x00u\x00l\x00t\x00C\x00o\x00n"
            b"\x00s\x00e\x00n\x00t\x00\x00"
        )
        test = win_lgpo._regexSearchKeyValueCombo(
            policy_data=policy_data,
            policy_regpath=policy_regpath,
            policy_regkey=policy_regkey,
        )
        self.assertIsNone(test)

    def test__regexSearchKeyValueCombo_disabled(self):
        """
        Make sure
        """
        policy_data = (
            b"[\x00s\x00o\x00f\x00t\x00w\x00a\x00r\x00e\x00\\\x00p"
            b"\x00o\x00l\x00i\x00c\x00i\x00e\x00s\x00\\\x00m\x00i"
            b"\x00c\x00r\x00o\x00s\x00o\x00f\x00t\x00\\\x00w\x00i"
            b"\x00n\x00d\x00o\x00w\x00s\x00\\\x00w\x00i\x00n\x00d"
            b"\x00o\x00w\x00s\x00 \x00e\x00r\x00r\x00o\x00r\x00 "
            b"\x00r\x00e\x00p\x00o\x00r\x00t\x00i\x00n\x00g\x00\\"
            b"\x00c\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00\x00;\x00*"
            b"\x00*\x00d\x00e\x00l\x00.\x00D\x00e\x00f\x00a\x00u"
            b"\x00l\x00t\x00C\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00"
            b"\x00;\x00\x01\x00\x00\x00;\x00\x04\x00\x00\x00;\x00 "
            b"\x00\x00\x00]\x00"
        )
        policy_regpath = (
            b"\x00s\x00o\x00f\x00t\x00w\x00a\x00r\x00e\x00\\\x00p"
            b"\x00o\x00l\x00i\x00c\x00i\x00e\x00s\x00\\\x00m\x00i"
            b"\x00c\x00r\x00o\x00s\x00o\x00f\x00t\x00\\\x00w\x00i"
            b"\x00n\x00d\x00o\x00w\x00s\x00\\\x00w\x00i\x00n\x00d"
            b"\x00o\x00w\x00s\x00 \x00e\x00r\x00r\x00o\x00r\x00 "
            b"\x00r\x00e\x00p\x00o\x00r\x00t\x00i\x00n\x00g\x00\\"
            b"\x00c\x00o\x00n\x00s\x00e\x00n\x00t\x00\x00"
        )
        policy_regkey = (
            b"\x00D\x00e\x00f\x00a\x00u\x00l\x00t\x00C\x00o\x00n"
            b"\x00s\x00e\x00n\x00t\x00\x00"
        )
        test = win_lgpo._regexSearchKeyValueCombo(
            policy_data=policy_data,
            policy_regpath=policy_regpath,
            policy_regkey=policy_regkey,
        )
        self.assertEqual(test, policy_data)

    def test__encode_string(self):
        """
        ``_encode_string`` should return a null terminated ``utf-16-le`` encoded
        string when a string value is passed
        """
        encoded_value = b"".join(
            ["Salt is awesome".encode("utf-16-le"), self.encoded_null]
        )
        value = win_lgpo._encode_string("Salt is awesome")
        self.assertEqual(value, encoded_value)

    def test__encode_string_empty_string(self):
        """
        ``_encode_string`` should return an encoded null when an empty string
        value is passed
        """
        value = win_lgpo._encode_string("")
        self.assertEqual(value, self.encoded_null)

    def test__encode_string_error(self):
        """
        ``_encode_string`` should raise an error if a non-string value is passed
        """
        self.assertRaises(TypeError, win_lgpo._encode_string, [1])
        test_list = ["item1", "item2"]
        self.assertRaises(TypeError, win_lgpo._encode_string, [test_list])
        test_dict = {"key1": "value1", "key2": "value2"}
        self.assertRaises(TypeError, win_lgpo._encode_string, [test_dict])

    def test__encode_string_none(self):
        """
        ``_encode_string`` should return an encoded null when ``None`` is passed
        """
        value = win_lgpo._encode_string(None)
        self.assertEqual(value, self.encoded_null)

    def test__multi_string_get_transform_list(self):
        """
        ``_multi_string_get_transform`` should return the list when a list is
        passed
        """
        test_value = ["Spongebob", "Squarepants"]
        value = win_lgpo._policy_info._multi_string_get_transform(item=test_value)
        self.assertEqual(value, test_value)

    def test__multi_string_get_transform_none(self):
        """
        ``_multi_string_get_transform`` should return "Not Defined" when
        ``None`` is passed
        """
        test_value = None
        value = win_lgpo._policy_info._multi_string_get_transform(item=test_value)
        self.assertEqual(value, "Not Defined")

    def test__multi_string_get_transform_invalid(self):
        """
        ``_multi_string_get_transform`` should return "Not Defined" when
        ``None`` is passed
        """
        test_value = "Some String"
        value = win_lgpo._policy_info._multi_string_get_transform(item=test_value)
        self.assertEqual(value, "Invalid Value")

    def test__multi_string_put_transform_list(self):
        """
        ``_multi_string_put_transform`` should return the list when a list is
        passed
        """
        test_value = ["Spongebob", "Squarepants"]
        value = win_lgpo._policy_info._multi_string_put_transform(item=test_value)
        self.assertEqual(value, test_value)

    def test__multi_string_put_transform_none(self):
        """
        ``_multi_string_put_transform`` should return ``None`` when
        "Not Defined" is passed
        """
        test_value = "Not Defined"
        value = win_lgpo._policy_info._multi_string_put_transform(item=test_value)
        self.assertEqual(value, None)

    def test__multi_string_put_transform_list_from_string(self):
        """
        ``_multi_string_put_transform`` should return a list when a comma
        delimited string is passed
        """
        test_value = "Spongebob,Squarepants"
        value = win_lgpo._policy_info._multi_string_put_transform(item=test_value)
        self.assertEqual(value, ["Spongebob", "Squarepants"])

    def test__multi_string_put_transform_invalid(self):
        """
        ``_multi_string_put_transform`` should return "Invalid" value if neither
        string nor list is passed
        """
        test_value = None
        value = win_lgpo._policy_info._multi_string_put_transform(item=test_value)
        self.assertEqual(value, "Invalid Value")


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLGPOGetPolicyADMXTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test functions related to the ``get_policy`` function using policy templates
    (admx/adml)
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def test_get_policy_name(self):
        result = win_lgpo.get_policy(
            policy_name="Allow Telemetry",
            policy_class="machine",
            return_value_only=True,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = "Not Configured"
        self.assertEqual(result, expected)

    def test_get_policy_id(self):
        result = win_lgpo.get_policy(
            policy_name="AllowTelemetry",
            policy_class="machine",
            return_value_only=True,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = "Not Configured"
        self.assertEqual(result, expected)

    def test_get_policy_name_full_return_full_names(self):
        result = win_lgpo.get_policy(
            policy_name="Allow Telemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {
            "Windows Components\\Data Collection and Preview Builds\\Allow Telemetry": (
                "Not Configured"
            )
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_full_names(self):
        result = win_lgpo.get_policy(
            policy_name="AllowTelemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {
            "Windows Components\\Data Collection and Preview Builds\\Allow Telemetry": (
                "Not Configured"
            )
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_name_full_return_ids(self):
        result = win_lgpo.get_policy(
            policy_name="Allow Telemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {"AllowTelemetry": "Not Configured"}
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_ids(self):
        result = win_lgpo.get_policy(
            policy_name="AllowTelemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {"AllowTelemetry": "Not Configured"}
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_ids_hierarchical(self):
        result = win_lgpo.get_policy(
            policy_name="AllowTelemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "WindowsComponents": {
                        "DataCollectionAndPreviewBuilds": {
                            "AllowTelemetry": "Not Configured"
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_name_return_full_names_hierarchical(self):
        result = win_lgpo.get_policy(
            policy_name="Allow Telemetry",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "Windows Components": {
                        "Data Collection and Preview Builds": {
                            "Allow Telemetry": "Not Configured"
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)

    @pytest.mark.destructive_test
    def test__load_policy_definitions(self):
        """
        Test that unexpected files in the PolicyDefinitions directory won't
        cause the _load_policy_definitions function to explode
        https://gitlab.com/saltstack/enterprise/lock/issues/3826
        """
        # The PolicyDefinitions directory should only contain ADMX files. We
        # want to make sure the `_load_policy_definitions` function skips non
        # ADMX files in this directory.
        # Create a bogus ADML file in PolicyDefinitions directory
        bogus_fle = os.path.join("c:\\Windows\\PolicyDefinitions", "_bogus.adml")
        cache_dir = os.path.join(win_lgpo.__opts__["cachedir"], "lgpo", "policy_defs")
        try:
            with salt.utils.files.fopen(bogus_fle, "w+") as fh:
                fh.write("<junk></junk>")
            # This function doesn't return anything (None), it just loads
            # the XPath structures into __context__. We're just making sure it
            # doesn't stack trace here
            self.assertIsNone(win_lgpo._load_policy_definitions())
        finally:
            # Remove source file
            os.remove(bogus_fle)
            # Remove cached file
            search_string = "{}\\_bogus*.adml".format(cache_dir)
            for file_name in glob.glob(search_string):
                os.remove(file_name)


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLGPOGetPolicyFromPolicyInfoTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test functions related to the ``get_policy`` function using _policy_info
    object
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def test_get_policy_name(self):
        result = win_lgpo.get_policy(
            policy_name="Network firewall: Public: Settings: Display a notification",
            policy_class="machine",
            return_value_only=True,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = "Not configured"
        self.assertEqual(result, expected)

    def test_get_policy_id(self):
        result = win_lgpo.get_policy(
            policy_name="WfwPublicSettingsNotification",
            policy_class="machine",
            return_value_only=True,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = "Not configured"
        self.assertEqual(result, expected)

    def test_get_policy_name_full_return(self):
        result = win_lgpo.get_policy(
            policy_name="Network firewall: Public: Settings: Display a notification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {
            "Network firewall: Public: Settings: Display a notification": (
                "Not configured"
            )
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return(self):
        result = win_lgpo.get_policy(
            policy_name="WfwPublicSettingsNotification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {
            "Network firewall: Public: Settings: Display a notification": (
                "Not configured"
            )
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_name_full_return_ids(self):
        result = win_lgpo.get_policy(
            policy_name="Network firewall: Public: Settings: Display a notification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {
            "Network firewall: Public: Settings: Display a notification": (
                "Not configured"
            )
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_ids(self):
        result = win_lgpo.get_policy(
            policy_name="WfwPublicSettingsNotification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {"WfwPublicSettingsNotification": "Not configured"}
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_ids_hierarchical(self):
        result = win_lgpo.get_policy(
            policy_name="WfwPublicSettingsNotification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=False,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Windows Settings": {
                    "Security Settings": {
                        "Windows Firewall with Advanced Security": {
                            "Windows Firewall with Advanced Security - Local Group Policy Object": {
                                "WfwPublicSettingsNotification": "Not configured"
                            }
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)

    def test_get_policy_id_full_return_full_names_hierarchical(self):
        result = win_lgpo.get_policy(
            policy_name="WfwPublicSettingsNotification",
            policy_class="machine",
            return_value_only=False,
            return_full_policy_names=True,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Windows Settings": {
                    "Security Settings": {
                        "Windows Firewall with Advanced Security": {
                            "Windows Firewall with Advanced Security - Local Group Policy Object": {
                                "Network firewall: Public: Settings: Display a notification": (
                                    "Not configured"
                                )
                            }
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLGPOPolicyInfoMechanismsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test getting local group policy settings defined in the _policy_info object
    Go through each mechanism
    """

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)
        cls.policy_data = salt.modules.win_lgpo._policy_info()

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = cls.policy_data = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def _test_policy(self, policy_name):
        """
        Helper function to get current setting
        """
        policy_definition = self.policy_data.policies["Machine"]["policies"][
            policy_name
        ]
        return salt.modules.win_lgpo._get_policy_info_setting(policy_definition)

    def test_registry_mechanism(self):
        """
        Test getting policy value using the Registry mechanism
        """
        policy_name = "RemoteRegistryExactPaths"
        result = self._test_policy(policy_name=policy_name)
        expected = [
            "System\\CurrentControlSet\\Control\\ProductOptions",
            "System\\CurrentControlSet\\Control\\Server Applications",
            "Software\\Microsoft\\Windows NT\\CurrentVersion",
        ]
        self.assertListEqual(result, expected)

    def test_secedit_mechanism(self):
        """
        Test getting policy value using the Secedit mechanism
        """
        policy_name = "LSAAnonymousNameLookup"
        result = self._test_policy(policy_name=policy_name)
        expected = "Disabled"
        self.assertEqual(result, expected)

    def test_netsh_mechanism(self):
        """
        Test getting the policy value using the NetSH mechanism
        """
        policy_name = "WfwDomainState"
        all_settings = {
            "State": "NotConfigured",
            "Inbound": "NotConfigured",
            "Outbound": "NotConfigured",
            "LocalFirewallRules": "NotConfigured",
            "LocalConSecRules": "NotConfigured",
            "InboundUserNotification": "NotConfigured",
            "RemoteManagement": "NotConfigured",
            "UnicastResponseToMulticast": "NotConfigured",
            "LogAllowedConnections": "NotConfigured",
            "LogDroppedConnections": "NotConfigured",
            "FileName": "NotConfigured",
            "MaxFileSize": "NotConfigured",
        }
        with patch(
            "salt.utils.win_lgpo_netsh.get_all_settings", return_value=all_settings
        ):
            result = self._test_policy(policy_name=policy_name)
        expected = "Not configured"
        self.assertEqual(result, expected)

    @pytest.mark.destructive_test
    def test_adv_audit_mechanism(self):
        """
        Test getting the policy value using the AdvAudit mechanism
        """
        system_root = os.environ.get("SystemRoot", "C:\\Windows")
        f_audit = os.path.join(system_root, "security", "audit", "audit.csv")
        f_audit_gpo = os.path.join(
            system_root,
            "System32",
            "GroupPolicy",
            "Machine",
            "Microsoft",
            "Windows NT",
            "Audit",
            "audit.csv",
        )
        if os.path.exists(f_audit):
            os.remove(f_audit)
        if os.path.exists(f_audit_gpo):
            os.remove(f_audit_gpo)
        policy_name = "AuditCredentialValidation"
        result = self._test_policy(policy_name=policy_name)
        expected = "Not Configured"
        self.assertEqual(result, expected)

    def test_net_user_modal_mechanism(self):
        """
        Test getting the policy value using the NetUserModal mechanism
        """
        policy_name = "PasswordHistory"
        result = self._test_policy(policy_name=policy_name)
        expected = 0
        self.assertEqual(result, expected)

    def test_lsa_rights_mechanism(self):
        """
        Test getting the policy value using the LsaRights mechanism
        """
        policy_name = "SeTrustedCredManAccessPrivilege"
        result = self._test_policy(policy_name=policy_name)
        expected = []
        self.assertEqual(result, expected)

    def test_script_ini_mechanism(self):
        """
        Test getting the policy value using the ScriptIni value
        """
        policy_name = "StartupScripts"
        result = self._test_policy(policy_name=policy_name)
        expected = None
        self.assertEqual(result, expected)


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
@pytest.mark.destructive_test
class WinLGPOGetPointAndPrintNCTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test variations of the Point and Print Restrictions policy when Not
    Configured (NC)
    """

    not_configured = False

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def setUp(self):
        if not self.not_configured:
            computer_policy = {"Point and Print Restrictions": "Not Configured"}
            win_lgpo.set_(computer_policy=computer_policy)
            self.not_configured = True

    def _get_policy_adm_setting(
        self, policy_name, policy_class, return_full_policy_names, hierarchical_return
    ):
        """
        Helper function to get current setting
        """
        # Get the policy
        success, policy_obj, _, _ = salt.modules.win_lgpo._lookup_admin_template(
            policy_name=policy_name, policy_class=policy_class, adml_language="en-US"
        )
        if success:
            return salt.modules.win_lgpo._get_policy_adm_setting(
                admx_policy=policy_obj,
                policy_class=policy_class,
                adml_language="en-US",
                return_full_policy_names=return_full_policy_names,
                hierarchical_return=hierarchical_return,
            )
        return "Policy Not Found"

    def test_point_and_print_not_configured(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {"PointAndPrint_Restrictions_Win7": "Not Configured"}
        self.assertDictEqual(result, expected)

    def test_point_and_print_not_configured_hierarchical(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=False,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "Printers": {"PointAndPrint_Restrictions_Win7": "Not Configured"}
                }
            }
        }
        self.assertDictEqual(result, expected)

    def test_point_and_print_not_configured_full_names(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {"Printers\\Point and Print Restrictions": "Not Configured"}
        self.assertDictEqual(result, expected)

    def test_point_and_print_not_configured_full_names_hierarchical(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=True,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "Printers": {"Point and Print Restrictions": "Not Configured"}
                }
            }
        }
        self.assertDictEqual(result, expected)


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
@pytest.mark.destructive_test
class WinLGPOGetPointAndPrintENTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test variations of the Point and Print Restrictions policy when Enabled (EN)
    """

    configured = False

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def setUp(self):
        if not self.configured:
            computer_policy = {
                "Point and Print Restrictions": {
                    "Users can only point and print to these servers": True,
                    "Enter fully qualified server names separated by semicolons": (
                        "fakeserver1;fakeserver2"
                    ),
                    "Users can only point and print to machines in their forest": True,
                    "When installing drivers for a new connection": (
                        "Show warning and elevation prompt"
                    ),
                    "When updating drivers for an existing connection": (
                        "Show warning only"
                    ),
                },
            }
            win_lgpo.set_(computer_policy=computer_policy)
            self.configured = True

    def _get_policy_adm_setting(
        self, policy_name, policy_class, return_full_policy_names, hierarchical_return
    ):
        """
        Helper function to get current setting
        """
        # Get the policy
        success, policy_obj, _, _ = salt.modules.win_lgpo._lookup_admin_template(
            policy_name=policy_name, policy_class=policy_class, adml_language="en-US"
        )
        if success:
            results = salt.modules.win_lgpo._get_policy_adm_setting(
                admx_policy=policy_obj,
                policy_class=policy_class,
                adml_language="en-US",
                return_full_policy_names=return_full_policy_names,
                hierarchical_return=hierarchical_return,
            )
            return results
        return "Policy Not Found"

    @pytest.mark.slow_test
    def test_point_and_print_enabled(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=False,
            hierarchical_return=False,
        )
        expected = {
            "PointAndPrint_Restrictions_Win7": {
                "PointAndPrint_NoWarningNoElevationOnInstall_Enum": (
                    "Show warning and elevation prompt"
                ),
                "PointAndPrint_NoWarningNoElevationOnUpdate_Enum": "Show warning only",
                "PointAndPrint_TrustedForest_Chk": True,
                "PointAndPrint_TrustedServers_Chk": True,
                "PointAndPrint_TrustedServers_Edit": "fakeserver1;fakeserver2",
            }
        }
        self.assertDictEqual(result, expected)

    def test_point_and_print_enabled_hierarchical(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=False,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "Printers": {
                        "PointAndPrint_Restrictions_Win7": {
                            "PointAndPrint_NoWarningNoElevationOnInstall_Enum": (
                                "Show warning and elevation prompt"
                            ),
                            "PointAndPrint_NoWarningNoElevationOnUpdate_Enum": (
                                "Show warning only"
                            ),
                            "PointAndPrint_TrustedForest_Chk": True,
                            "PointAndPrint_TrustedServers_Chk": True,
                            "PointAndPrint_TrustedServers_Edit": (
                                "fakeserver1;fakeserver2"
                            ),
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)

    def test_point_and_print_enabled_full_names(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=True,
            hierarchical_return=False,
        )
        expected = {
            "Printers\\Point and Print Restrictions": {
                "Enter fully qualified server names separated by semicolons": (
                    "fakeserver1;fakeserver2"
                ),
                "When installing drivers for a new connection": (
                    "Show warning and elevation prompt"
                ),
                "Users can only point and print to machines in their forest": True,
                "Users can only point and print to these servers": True,
                "When updating drivers for an existing connection": "Show warning only",
            }
        }
        self.assertDictEqual(result, expected)

    @pytest.mark.slow_test
    def test_point_and_print_enabled_full_names_hierarchical(self):
        result = self._get_policy_adm_setting(
            policy_name="Point and Print Restrictions",
            policy_class="Machine",
            return_full_policy_names=True,
            hierarchical_return=True,
        )
        expected = {
            "Computer Configuration": {
                "Administrative Templates": {
                    "Printers": {
                        "Point and Print Restrictions": {
                            "Enter fully qualified server names separated by semicolons": (
                                "fakeserver1;fakeserver2"
                            ),
                            "When installing drivers for a new connection": (
                                "Show warning and elevation prompt"
                            ),
                            "Users can only point and print to machines in their forest": True,
                            "Users can only point and print to these servers": True,
                            "When updating drivers for an existing connection": (
                                "Show warning only"
                            ),
                        }
                    }
                }
            }
        }
        self.assertDictEqual(result, expected)


@skipIf(not salt.utils.platform.is_windows(), "System is not Windows")
class WinLGPOGetPolicyFromPolicyResources(TestCase, LoaderModuleMockMixin):
    """
    Test functions related to policy info gathered from ADMX/ADML files
    """

    adml_data = None

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.utils = salt.loader.utils(cls.opts)
        cls.modules = salt.loader.minion_mods(cls.opts, utils=cls.utils)

    @classmethod
    def tearDownClass(cls):
        cls.opts = cls.utils = cls.modules = None

    def setup_loader_modules(self):
        return {
            win_lgpo: {
                "__opts__": copy.deepcopy(self.opts),
                "__salt__": self.modules,
                "__utils__": self.utils,
            }
        }

    def setUp(self):
        if self.adml_data is None:
            self.adml_data = win_lgpo._get_policy_resources(language="en-US")

    def test__getAdmlPresentationRefId(self):
        ref_id = "LetAppsAccessAccountInfo_Enum"
        expected = "Default for all apps"
        result = win_lgpo._getAdmlPresentationRefId(self.adml_data, ref_id)
        self.assertEqual(result, expected)

    def test__getAdmlPresentationRefId_result_text_is_none(self):
        ref_id = "LetAppsAccessAccountInfo_UserInControlOfTheseApps_List"
        expected = (
            "Put user in control of these specific apps (use Package Family Names)"
        )
        result = win_lgpo._getAdmlPresentationRefId(self.adml_data, ref_id)
        self.assertEqual(result, expected)



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>
"""


import salt.modules.osquery as osquery
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class OSQueryTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.iptables
    """

    def setup_loader_modules(self):
        return {osquery: {}}

    def test_version(self):
        """
        Test the version returned from OSQuery
        """
        _table_attrs_results = [
            "pid",
            "uuid",
            "instance_id",
            "version",
            "config_hash",
            "config_valid",
            "extensions",
            "build_platform",
            "build_distro",
            "start_time",
            "watcher",
        ]

        _os_query_results = {"data": [{"version": "2.6.1"}], "result": True}

        with patch.object(
            osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
        ):
            with patch.object(
                osquery, "_osquery", MagicMock(return_value=_os_query_results)
            ):
                self.assertEqual(osquery.version(), "2.6.1")

    def test_deb_packages(self):
        """
        Test the results returned from the deb_packages function
        """
        _os_query_results = {
            "data": [
                {
                    "arch": "amd64",
                    "name": "accountsservice",
                    "revision": "1",
                    "size": "451",
                    "source": "",
                    "version": "0.6.45-1",
                },
                {
                    "arch": "amd64",
                    "name": "acetoneiso",
                    "revision": "2+b2",
                    "size": "1820",
                    "source": "acetoneiso (2.4-2)",
                    "version": "2.4-2+b2",
                },
                {
                    "arch": "amd64",
                    "name": "acl",
                    "revision": "3+b1",
                    "size": "200",
                    "source": "acl (2.2.52-3)",
                    "version": "2.2.52-3+b1",
                },
                {
                    "arch": "amd64",
                    "name": "adb",
                    "revision": "2",
                    "size": "189",
                    "source": "android-platform-system-core",
                    "version": "1: 7.0.0+r33-2",
                },
            ],
            "result": True,
        }
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                self.assertEqual(osquery.deb_packages(), _os_query_results)

    def test_deb_packages_with_attrs(self):
        """
        Test the results returned from the deb_packages function
        with attributes
        """
        _table_attrs_results = ["name", "version", "source", "size", "arch", "revision"]

        _os_query_results = {
            "data": [
                {"name": "accountsservice", "version": "0.6.45-1"},
                {"name": "acetoneiso", "version": "2.4-2+b2"},
                {"name": "acl", "version": "2.2.52-3+b1"},
                {"name": "adb", "version": "1: 7.0.0+r33-2"},
            ],
            "result": True,
        }

        with patch.object(
            osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
        ):
            with patch.object(
                osquery, "_osquery", MagicMock(return_value=_os_query_results)
            ):
                with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                    self.assertEqual(
                        osquery.deb_packages(attrs=["name", "version"]),
                        _os_query_results,
                    )

    def test_kernel_modules(self):
        """
        Test the results returned from the kernel_modules function
        """
        _os_query_results = {
            "data": [
                {
                    "address": "0xffffffffc14f2000",
                    "name": "nls_utf8",
                    "size": "16384",
                    "status": "Live",
                    "used_by": "-",
                },
                {
                    "address": "0xffffffffc1599000",
                    "name": "udf",
                    "size": "90112",
                    "status": "Live",
                    "used_by": "-",
                },
                {
                    "address": "0xffffffffc14b5000",
                    "name": "crc_itu_t",
                    "size": "16384",
                    "status": "Live",
                    "used_by": "udf",
                },
            ],
            "result": True,
        }
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                self.assertEqual(osquery.kernel_modules(), _os_query_results)

    def test_kernel_modules_with_attrs(self):
        """
        Test the results returned from the kernel_modules function
        with attributes
        """
        _table_attrs_results = ["address", "name", "size", "status", "used_by"]

        _os_query_results = {
            "data": [
                {"name": "nls_utf8", "status": "Live"},
                {"name": "udf", "status": "Live"},
                {"name": "crc_itu_t", "status": "Live"},
            ],
            "result": True,
        }
        with patch.object(
            osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
        ):
            with patch.object(
                osquery, "_osquery", MagicMock(return_value=_os_query_results)
            ):
                with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                    self.assertEqual(
                        osquery.kernel_modules(attrs=["name", "status"]),
                        _os_query_results,
                    )

    def test_osquery_info(self):
        """
        Test the results returned from the kernel_modules function
        with attributes
        """
        _table_attrs_results = [
            "pid",
            "uuid",
            "instance_id",
            "version",
            "config_hash",
            "config_valid",
            "extensions",
            "build_platform",
            "build_distro",
            "start_time",
            "watcher",
        ]

        _os_query_results = {
            "data": [
                {
                    "build_platform": "ubuntu",
                    "start_time": "1514484833",
                    "uuid": "D31FD400-7277-11E3-ABA6-B8AEED7E173B",
                    "build_distro": "xenial",
                    "pid": "24288",
                    "watcher": "-1",
                    "instance_id": "dff196b0-5c91-4105-962b-28660d7aa282",
                    "version": "2.6.1",
                    "extensions": "inactive",
                    "config_valid": "0",
                    "config_hash": "",
                }
            ],
            "result": True,
        }

        with patch.object(
            osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
        ):
            with patch.object(
                osquery, "_osquery", MagicMock(return_value=_os_query_results)
            ):
                with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                    self.assertEqual(osquery.osquery_info(), _os_query_results)

    def test_osquery_info_with_attrs(self):
        """
        Test the results returned from the kernel_modules function
        with attributes
        """
        _table_attrs_results = [
            "pid",
            "uuid",
            "instance_id",
            "version",
            "config_hash",
            "config_valid",
            "extensions",
            "build_platform",
            "build_distro",
            "start_time",
            "watcher",
        ]

        _os_query_results = {
            "data": [{"build_platform": "ubuntu", "start_time": "1514484833"}],
            "result": True,
        }

        with patch.object(
            osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
        ):
            with patch.object(
                osquery, "_osquery", MagicMock(return_value=_os_query_results)
            ):
                with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                    self.assertEqual(
                        osquery.osquery_info(attrs=["build_platform", "start_time"]),
                        _os_query_results,
                    )



import pytest
import salt.states.file as file
import salt.utils.win_dacl as win_dacl

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {
        file: {"__opts__": {"test": False}},
    }


def test__check_directory_win_owner(tmp_path):
    path = str(tmp_path)
    _, comment, changes = file._check_directory_win(name=path, win_owner="Everyone")
    assert path in comment
    assert changes == {"owner": "Everyone"}


def test__check_directory_win_grant_perms_basic(tmp_path):
    path = str(tmp_path)
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": "full_control",
        }
    }
    expected = {
        "grant_perms": {
            "Guest": {
                "applies_to": "this_folder_subfolders_files",
                "permissions": "full_control",
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_grant_perms_basic_existing_user(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions=["write_data", "write_attributes"],
        access_mode="grant",
    )
    perms = {"Guest": {"perms": "full_control"}}
    expected = {"grant_perms": {"Guest": {"permissions": "full_control"}}}
    _, comment, changes = file._check_directory_win(name=path, win_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_grant_perms_advanced(tmp_path):
    path = str(tmp_path)
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": ["read_data", "write_data", "create_files"],
        }
    }
    expected = {
        "grant_perms": {
            "Guest": {
                "applies_to": "this_folder_subfolders_files",
                "permissions": ["read_data", "write_data", "create_files"],
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_grant_perms_advanced_existing_user(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions="full_control",
        access_mode="grant",
    )
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": ["read_data", "write_data", "create_files"],
        }
    }
    expected = {
        "grant_perms": {
            "Guest": {"permissions": ["read_data", "write_data", "create_files"]}
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_grant_perms_basic_no_applies_to(tmp_path):
    path = str(tmp_path)
    perms = {"Guest": {"perms": "full_control"}}
    expected = {"grant_perms": {"Guest": {"permissions": "full_control"}}}
    _, comment, changes = file._check_directory_win(name=path, win_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_deny_perms_basic(tmp_path):
    path = str(tmp_path)
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": "full_control",
        }
    }
    expected = {
        "deny_perms": {
            "Guest": {
                "applies_to": "this_folder_subfolders_files",
                "permissions": "full_control",
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_deny_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_deny_perms_basic_existing_user(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions=["write_data", "write_attributes"],
        access_mode="deny",
    )
    perms = {"Guest": {"perms": "full_control"}}
    expected = {"deny_perms": {"Guest": {"permissions": "full_control"}}}
    _, comment, changes = file._check_directory_win(name=path, win_deny_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_deny_perms_advanced(tmp_path):
    path = str(tmp_path)
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": ["read_data", "write_data", "create_files"],
        }
    }
    expected = {
        "deny_perms": {
            "Guest": {
                "applies_to": "this_folder_subfolders_files",
                "permissions": ["read_data", "write_data", "create_files"],
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_deny_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_deny_perms_advanced_existing_user(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions="full_control",
        access_mode="deny",
    )
    perms = {
        "Guest": {
            "applies_to": "this_folder_subfolders_files",
            "perms": ["read_data", "write_data", "create_files"],
        }
    }
    expected = {
        "deny_perms": {
            "Guest": {"permissions": ["read_data", "write_data", "create_files"]}
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_deny_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_deny_perms_basic_no_applies_to(tmp_path):
    path = str(tmp_path)
    perms = {"Guest": {"perms": "full_control"}}
    expected = {"deny_perms": {"Guest": {"permissions": "full_control"}}}
    _, comment, changes = file._check_directory_win(name=path, win_deny_perms=perms)
    assert path in comment
    assert changes == expected


def test__check_directory_win_inheritance(tmp_path):
    path = str(tmp_path)
    expected = {}
    _, comment, changes = file._check_directory_win(name=path, win_inheritance=True)
    assert path in comment
    assert changes == expected


def test__check_directory_win_inheritance_false(tmp_path):
    path = str(tmp_path)
    expected = {"inheritance": False}
    _, comment, changes = file._check_directory_win(name=path, win_inheritance=False)
    assert path in comment
    assert changes == expected


def test__check_directory_reset_no_non_inherited_users(tmp_path):
    path = str(tmp_path)
    expected = {}
    _, comment, changes = file._check_directory_win(name=path, win_perms_reset=True)
    assert path in comment
    assert changes == expected


def test__check_directory_reset_non_inherited_users_grant(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions="full_control",
        access_mode="grant",
        reset_perms=True,
    )
    expected = {
        "remove_perms": {
            "Guest": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_perms_reset=True)
    assert path in comment
    assert changes == expected


def test__check_directory_reset_non_inherited_users_deny(tmp_path):
    path = str(tmp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions="full_control",
        access_mode="deny",
        reset_perms=True,
    )
    expected = {
        "remove_perms": {
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            }
        }
    }
    _, comment, changes = file._check_directory_win(name=path, win_perms_reset=True)
    assert path in comment
    assert changes == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.sysbench as sysbench
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SysbenchTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases to salt.modules.sysbench
    """

    def setup_loader_modules(self):
        return {sysbench: {}}

    def test_cpu(self):
        """
        Test to tests to the CPU performance of minions.
        """
        with patch.dict(
            sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}
        ):
            with patch.object(sysbench, "_parser", return_value={"A": "a"}):
                self.assertEqual(
                    sysbench.cpu(),
                    {
                        "Prime numbers limit: 500": {"A": "a"},
                        "Prime numbers limit: 5000": {"A": "a"},
                        "Prime numbers limit: 2500": {"A": "a"},
                        "Prime numbers limit: 1000": {"A": "a"},
                    },
                )

    def test_threads(self):
        """
        Test to this tests the performance of the processor's scheduler
        """
        with patch.dict(
            sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}
        ):
            with patch.object(sysbench, "_parser", return_value={"A": "a"}):
                self.assertEqual(
                    sysbench.threads(),
                    {
                        "Yields: 500 Locks: 8": {"A": "a"},
                        "Yields: 200 Locks: 4": {"A": "a"},
                        "Yields: 1000 Locks: 16": {"A": "a"},
                        "Yields: 100 Locks: 2": {"A": "a"},
                    },
                )

    def test_mutex(self):
        """
        Test to tests the implementation of mutex
        """
        with patch.dict(
            sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}
        ):
            with patch.object(sysbench, "_parser", return_value={"A": "a"}):
                self.assertEqual(
                    sysbench.mutex(),
                    {
                        "Mutex: 1000 Locks: 25000 Loops: 10000": {"A": "a"},
                        "Mutex: 50 Locks: 10000 Loops: 2500": {"A": "a"},
                        "Mutex: 1000 Locks: 10000 Loops: 5000": {"A": "a"},
                        "Mutex: 500 Locks: 50000 Loops: 5000": {"A": "a"},
                        "Mutex: 500 Locks: 25000 Loops: 2500": {"A": "a"},
                        "Mutex: 500 Locks: 10000 Loops: 10000": {"A": "a"},
                        "Mutex: 50 Locks: 50000 Loops: 10000": {"A": "a"},
                        "Mutex: 1000 Locks: 50000 Loops: 2500": {"A": "a"},
                        "Mutex: 50 Locks: 25000 Loops: 5000": {"A": "a"},
                    },
                )

    def test_memory(self):
        """
        Test to this tests the memory for read and write operations.
        """
        with patch.dict(
            sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}
        ):
            with patch.object(sysbench, "_parser", return_value={"A": "a"}):
                self.assertEqual(
                    sysbench.memory(),
                    {
                        "Operation: read Scope: local": {"A": "a"},
                        "Operation: write Scope: local": {"A": "a"},
                        "Operation: read Scope: global": {"A": "a"},
                        "Operation: write Scope: global": {"A": "a"},
                    },
                )

    def test_fileio(self):
        """
        Test to this tests for the file read and write operations
        """
        with patch.dict(
            sysbench.__salt__, {"cmd.run": MagicMock(return_value={"A": "a"})}
        ):
            with patch.object(sysbench, "_parser", return_value={"A": "a"}):
                self.assertEqual(
                    sysbench.fileio(),
                    {
                        "Mode: seqrd": {"A": "a"},
                        "Mode: seqwr": {"A": "a"},
                        "Mode: rndrd": {"A": "a"},
                        "Mode: rndwr": {"A": "a"},
                        "Mode: seqrewr": {"A": "a"},
                        "Mode: rndrw": {"A": "a"},
                    },
                )

    def test_ping(self):
        """
        Test to ping
        """
        self.assertTrue(sysbench.ping())



"""
Capirca ACL
===========

Generate ACL (firewall) configuration for network devices.

.. versionadded:: 2017.7.0

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Robert Ankeny <robankeny@google.com>
:maturity:   new
:depends:    capirca
:platform:   unix

Dependencies
------------

The firewall configuration is generated by Capirca_.

.. _Capirca: https://github.com/google/capirca

To install Capirca, execute: ``pip install capirca``.
"""

import datetime
import inspect
import logging
import re

import salt.utils.files

log = logging.getLogger(__file__)


try:
    import capirca
    import capirca.aclgen
    import capirca.lib.policy
    import capirca.lib.aclgenerator

    HAS_CAPIRCA = True
except ImportError:
    HAS_CAPIRCA = False


# ------------------------------------------------------------------------------
# module properties
# ------------------------------------------------------------------------------

__virtualname__ = "capirca"
__proxyenabled__ = ["*"]
# allow any proxy type

# ------------------------------------------------------------------------------
# property functions
# ------------------------------------------------------------------------------


def __virtual__():
    """
    This module requires at least Capirca to work.
    """
    if HAS_CAPIRCA:
        return __virtualname__
    else:
        return (False, "The capirca module (capirca_acl) cannot be loaded.")


# ------------------------------------------------------------------------------
# module globals
# ------------------------------------------------------------------------------


# define the default values for all possible term fields
# we could also extract them from the `policy` module, inspecting the `Policy`
# class, but that might be overkill & it would make the code less obvious.
# we can revisit this later if necessary.

_TERM_FIELDS = {
    "action": [],
    "address": [],
    "address_exclude": [],
    "comment": [],
    "counter": None,
    "expiration": None,
    "destination_address": [],
    "destination_address_exclude": [],
    "destination_port": [],
    "destination_prefix": [],
    "forwarding_class": [],
    "forwarding_class_except": [],
    "logging": [],
    "log_name": None,
    "loss_priority": None,
    "option": [],
    "owner": None,
    "policer": None,
    "port": [],
    "precedence": [],
    "principals": [],
    "protocol": [],
    "protocol_except": [],
    "qos": None,
    "pan_application": [],
    "routing_instance": None,
    "source_address": [],
    "source_address_exclude": [],
    "source_port": [],
    "source_prefix": [],
    "verbatim": [],
    "packet_length": None,
    "fragment_offset": None,
    "hop_limit": None,
    "icmp_type": [],
    "icmp_code": None,
    "ether_type": [],
    "traffic_class_count": None,
    "traffic_type": [],
    "translated": False,
    "dscp_set": None,
    "dscp_match": [],
    "dscp_except": [],
    "next_ip": None,
    "flexible_match_range": [],
    "source_prefix_except": [],
    "destination_prefix_except": [],
    "vpn": None,
    "source_tag": [],
    "destination_tag": [],
    "source_interface": None,
    "destination_interface": None,
    "platform": [],
    "platform_exclude": [],
    "timeout": None,
    "flattened": False,
    "flattened_addr": None,
    "flattened_saddr": None,
    "flattened_daddr": None,
    "priority": None,
    "ttl": None,
}

# IP-type fields
# when it comes to IP fields, Capirca does not ingest raw text
# but they need to be converted to `nacaddr.IP`
# this pre-processing is done in `_clean_term_opts`
_IP_FILEDS = [
    "source_address",
    "source_address_exclude",
    "destination_address",
    "address",
    "address_exclude",
    "flattened_addr",
    "flattened_saddr",
    "flattened_daddr",
    "next_ip",
]

_SERVICES = {}

# ------------------------------------------------------------------------------
# helper functions -- will not be exported
# ------------------------------------------------------------------------------


if HAS_CAPIRCA:
    _TempTerm = capirca.lib.policy.Term

    def _add_object(self, obj):
        return

    setattr(_TempTerm, "AddObject", _add_object)
    dumy_term = _TempTerm(None)
    for item in dir(dumy_term):
        if hasattr(item, "__func__") or item.startswith("_") or item != item.lower():
            continue
        _TERM_FIELDS[item] = getattr(dumy_term, item)

    class _Policy(capirca.lib.policy.Policy):
        """
        Extending the Capirca Policy class to allow inserting custom filters.
        """

        def __init__(self):
            self.filters = []
            self.filename = ""

    class _Term(capirca.lib.policy.Term):
        """
        Extending the Capirca Term class to allow setting field valued on the fly.
        """

        def __init__(self):
            for field, default in _TERM_FIELDS.items():
                setattr(self, field, default)


def _import_platform_generator(platform):
    """
    Given a specific platform (under the Capirca conventions),
    return the generator class.
    The generator class is identified looking under the <platform> module
    for a class inheriting the `ACLGenerator` class.
    """
    log.debug("Using platform: %s", platform)
    for mod_name, mod_obj in inspect.getmembers(capirca.aclgen):
        if mod_name == platform and inspect.ismodule(mod_obj):
            for plat_obj_name, plat_obj in inspect.getmembers(
                mod_obj
            ):  # pylint: disable=unused-variable
                if inspect.isclass(plat_obj) and issubclass(
                    plat_obj, capirca.lib.aclgenerator.ACLGenerator
                ):
                    log.debug("Identified Capirca class %s for %s", plat_obj, platform)
                    return plat_obj
    log.error("Unable to identify any Capirca plaform class for %s", platform)


def _get_services_mapping():
    """
    Build a map of services based on the IANA assignment list:
    http://www.iana.org/assignments/port-numbers

    It will load the /etc/services file and will build the mapping on the fly,
    similar to the Capirca's SERVICES file:
    https://github.com/google/capirca/blob/master/def/SERVICES.svc

    As this module is be available on Unix systems only,
    we'll read the services from /etc/services.
    In the worst case, the user will not be able to specify the
    services shortcut and they will need to specify the protocol / port combination
    using the source_port / destination_port & protocol fields.
    """
    if _SERVICES:
        return _SERVICES
    services_txt = ""
    try:
        with salt.utils.files.fopen("/etc/services", "r") as srv_f:
            services_txt = salt.utils.stringutils.to_unicode(srv_f.read())
    except OSError as ioe:
        log.error("Unable to read from /etc/services:")
        log.error(ioe)
        return _SERVICES  # no mapping possible, sorry
        # will return the default mapping
    service_rgx = re.compile(r"^([a-zA-Z0-9-]+)\s+(\d+)\/(tcp|udp)(.*)$")
    for line in services_txt.splitlines():
        service_rgx_s = service_rgx.search(line)
        if service_rgx_s and len(service_rgx_s.groups()) == 4:
            srv_name, port, protocol, _ = service_rgx_s.groups()
            if srv_name not in _SERVICES:
                _SERVICES[srv_name] = {"port": [], "protocol": []}
            try:
                _SERVICES[srv_name]["port"].append(int(port))
            except ValueError as verr:
                log.error(verr)
                log.error("Did not read that properly:")
                log.error(line)
                log.error(
                    "Please report the above error: %s does not seem a valid port"
                    " value!",
                    port,
                )
            _SERVICES[srv_name]["protocol"].append(protocol)
    return _SERVICES


def _translate_port(port):
    """
    Look into services and return the port value using the
    service name as lookup value.
    """
    services = _get_services_mapping()
    if port in services and services[port]["port"]:
        return services[port]["port"][0]
    return port


def _make_it_list(dict_, field_name, value):
    """
    Return the object list.
    """
    prev_value = []
    # firsly we'll collect the prev value
    if field_name in dict_:
        prev_value = dict_[field_name]
    if value is None:
        return prev_value
    elif isinstance(value, (tuple, list)):
        # other type of iterables
        if field_name in ("source_port", "destination_port"):
            # port fields are more special
            # they can either be a list of integers, either a list of tuples
            # list of integers = a list of ports
            # list of tuples = a list of ranges,
            # e.g.: [(1000, 2000), (3000, 4000)] means the 1000-2000 and 3000-4000 ranges
            portval = []
            for port in value:
                if not isinstance(port, (tuple, list)):
                    # to make sure everything is consistent,
                    # we'll transform indivitual ports into tuples
                    # thus an individual port e.g. 1000 will be transormed into the port range 1000-1000
                    # which is the equivalent
                    # but assures consistency for the Capirca parser
                    portval.append((port, port))
                else:
                    portval.append(port)
            translated_portval = []
            # and the ports sent as string, e.g. ntp instead of 123
            # needs to be translated
            # again, using the same /etc/services
            for port_start, port_end in portval:
                if not isinstance(port_start, int):
                    port_start = _translate_port(port_start)
                if not isinstance(port_end, int):
                    port_end = _translate_port(port_end)
                translated_portval.append((port_start, port_end))
            return list(set(prev_value + translated_portval))
        return list(set(prev_value + list(value)))
    if field_name in ("source_port", "destination_port"):
        if not isinstance(value, int):
            value = _translate_port(value)
        return list(set(prev_value + [(value, value)]))  # a list of tuples
    # anything else will be enclosed in a list-type
    return list(set(prev_value + [value]))


def _clean_term_opts(term_opts):
    """
    Cleanup the term opts:

    - strip Null and empty valuee, defaulting their value to their base definition from _TERM_FIELDS
    - convert to `nacaddr.IP` fields from `_IP_FILEDS`
    - create lists for those fields requiring it
    """
    clean_opts = {}
    _services = _get_services_mapping()
    for field, value in term_opts.items():
        # firstly we'll process special fields like source_service or destination_services
        # which will inject values directly in the source or destination port and protocol
        if field == "source_service" and value:
            if isinstance(value, str):
                value = _make_it_list(clean_opts, field, value)
            log.debug("Processing special source services:")
            log.debug(value)
            for service in value:
                if service and service in _services:
                    # if valid source_service
                    # take the port and protocol values from the global and inject in the term config
                    clean_opts["source_port"] = _make_it_list(
                        clean_opts, "source_port", _services[service]["port"]
                    )
                    clean_opts["protocol"] = _make_it_list(
                        clean_opts, "protocol", _services[service]["protocol"]
                    )
            log.debug(
                "Built source_port field, after processing special source services:"
            )
            log.debug(clean_opts.get("source_port"))
            log.debug("Built protocol field, after processing special source services:")
            log.debug(clean_opts.get("protocol"))
        elif field == "destination_service" and value:
            if isinstance(value, str):
                value = _make_it_list(clean_opts, field, value)
            log.debug("Processing special destination services:")
            log.debug(value)
            for service in value:
                if service and service in _services:
                    # if valid destination_service
                    # take the port and protocol values from the global and inject in the term config
                    clean_opts["destination_port"] = _make_it_list(
                        clean_opts, "destination_port", _services[service]["port"]
                    )
                    clean_opts["protocol"] = _make_it_list(
                        clean_opts, "protocol", _services[service]["protocol"]
                    )
            log.debug(
                "Built source_port field, after processing special destination"
                " services:"
            )
            log.debug(clean_opts.get("destination_service"))
            log.debug(
                "Built protocol field, after processing special destination services:"
            )
            log.debug(clean_opts.get("protocol"))
        # not a special field, but it has to be a valid one
        elif field in _TERM_FIELDS and value and value != _TERM_FIELDS[field]:
            # if not a special field type
            if isinstance(_TERM_FIELDS[field], list):
                value = _make_it_list(clean_opts, field, value)
            if field in _IP_FILEDS:
                # IP-type fields need to be transformed
                ip_values = []
                for addr in value:
                    ip_values.append(capirca.lib.policy.nacaddr.IP(addr))
                value = ip_values[:]
            clean_opts[field] = value
    return clean_opts


def _lookup_element(lst, key):
    """
    Find an dictionary in a list of dictionaries, given its main key.
    """
    if not lst:
        return {}
    for ele in lst:
        if not ele or not isinstance(ele, dict):
            continue
        if key in ele:
            return ele[key]
    return {}


def _get_pillar_cfg(pillar_key, pillarenv=None, saltenv=None):
    """
    Retrieve the pillar data from the right environment.
    """
    pillar_cfg = __salt__["pillar.get"](
        pillar_key, pillarenv=pillarenv, saltenv=saltenv
    )
    return pillar_cfg


def _cleanup(lst):
    """
    Return a list of non-empty dictionaries.
    """
    clean = []
    for ele in lst:
        if ele and isinstance(ele, dict):
            clean.append(ele)
    return clean


def _merge_list_of_dict(first, second, prepend=True):
    """
    Merge lists of dictionaries.
    Each element of the list is a dictionary having one single key.
    That key is then used as unique lookup.
    The first element list has higher priority than the second.
    When there's an overlap between the two lists,
    it won't change the position, but the content.
    """
    first = _cleanup(first)
    second = _cleanup(second)
    if not first and not second:
        return []
    if not first and second:
        return second
    if first and not second:
        return first
    # Determine overlaps
    # So we don't change the position of the existing terms/filters
    overlaps = []
    merged = []
    appended = []
    for ele in first:
        if _lookup_element(second, next(iter(ele))):
            overlaps.append(ele)
        elif prepend:
            merged.append(ele)
        elif not prepend:
            appended.append(ele)
    for ele in second:
        ele_key = next(iter(ele))
        if _lookup_element(overlaps, ele_key):
            # If there's an overlap, get the value from the first
            # But inserted into the right position
            ele_val_first = _lookup_element(first, ele_key)
            merged.append({ele_key: ele_val_first})
        else:
            merged.append(ele)
    if not prepend:
        merged.extend(appended)
    return merged


def _get_term_object(
    filter_name,
    term_name,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=True,
    **term_fields
):
    """
    Return an instance of the ``_Term`` class given the term options.
    """
    log.debug("Generating config for term %s under filter %s", term_name, filter_name)
    term = _Term()
    term.name = term_name
    term_opts = {}
    if merge_pillar:
        term_opts = get_term_pillar(
            filter_name,
            term_name,
            pillar_key=pillar_key,
            saltenv=saltenv,
            pillarenv=pillarenv,
        )
        log.debug("Merging with pillar data:")
        log.debug(term_opts)
        term_opts = _clean_term_opts(term_opts)
        log.debug("Cleaning up pillar data:")
        log.debug(term_opts)
    log.debug("Received processing opts:")
    log.debug(term_fields)
    log.debug("Cleaning up processing opts:")
    term_fields = _clean_term_opts(term_fields)
    log.debug(term_fields)
    log.debug("Final term opts:")
    term_opts.update(term_fields)
    log.debug(term_fields)
    for field, value in term_opts.items():
        # setting the field attributes to the term instance of _Term
        setattr(term, field, value)
    log.debug("Term config:")
    log.debug(str(term))
    return term


def _get_policy_object(
    platform,
    filters=None,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=True,
):
    """
    Return an instance of the ``_Policy`` class given the filters config.
    """
    policy = _Policy()
    policy_filters = []
    if not filters:
        filters = []
    for filter_ in filters:
        if not filter_ or not isinstance(filter_, dict):
            continue  # go to the next filter
        filter_name, filter_config = next(iter(filter_.items()))
        header = capirca.lib.policy.Header()  # same header everywhere
        target_opts = [platform, filter_name]
        filter_options = filter_config.pop("options", None)
        if filter_options:
            filter_options = _make_it_list({}, filter_name, filter_options)
            # make sure the filter options are sent as list
            target_opts.extend(filter_options)
        target = capirca.lib.policy.Target(target_opts)
        header.AddObject(target)
        filter_terms = []
        for term_ in filter_config.get("terms", []):
            if term_ and isinstance(term_, dict):
                term_name, term_fields = next(iter(term_.items()))
                term = _get_term_object(
                    filter_name,
                    term_name,
                    pillar_key=pillar_key,
                    pillarenv=pillarenv,
                    saltenv=saltenv,
                    merge_pillar=merge_pillar,
                    **term_fields
                )
            filter_terms.append(term)
        policy_filters.append((header, filter_terms))
    policy.filters = policy_filters
    log.debug("Policy config:")
    log.debug(str(policy))
    platform_generator = _import_platform_generator(platform)
    policy_config = platform_generator(policy, 2)
    log.debug("Generating policy config for %s:", platform)
    log.debug(str(policy_config))
    return policy_config


def _revision_tag(
    text,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
):
    """
    Refactor revision tag comments.
    Capirca generates the filter text having the following tag keys:

    - $Id:$
    - $Revision:$
    - $Date:$

    This function goes through all the config lines and replaces
    those tags with the content requested by the user.
    If a certain value is not provided, the corresponding tag will be stripped.
    """
    timestamp = datetime.datetime.now().strftime(revision_date_format)
    new_text = []
    for line in text.splitlines():
        if "$Id:$" in line:
            if not revision_id:  # if no explicit revision ID required
                continue  # jump to next line, ignore this one
            line = line.replace("$Id:$", "$Id: {rev_id} $".format(rev_id=revision_id))
        if "$Revision:$" in line:
            if not revision_no:  # if no explicit revision number required
                continue  # jump to next line, ignore this one
            line = line.replace(
                "$Revision:$", "$Revision: {rev_no} $".format(rev_no=revision_no)
            )
        if "$Date:$" in line:
            if not revision_date:
                continue  # jump
            line = line.replace("$Date:$", "$Date: {ts} $".format(ts=timestamp))
        new_text.append(line)
    return "\n".join(new_text)


# ------------------------------------------------------------------------------
# callable functions
# ------------------------------------------------------------------------------


def get_term_config(
    platform,
    filter_name,
    term_name,
    filter_options=None,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=True,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
    source_service=None,
    destination_service=None,
    **term_fields
):
    """
    Return the configuration of a single policy term.

    platform
        The name of the Capirca platform.

    filter_name
        The name of the policy filter.

    term_name
        The name of the term.

    filter_options
        Additional filter options. These options are platform-specific.
        E.g.: ``inet6``, ``bridge``, ``object-group``,
        See the complete list of options_.

        .. _options: https://github.com/google/capirca/wiki/Policy-format#header-section

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.
        If the pillar contains the following structure:

        .. code-block:: yaml

            firewall:
              - my-filter:
                  terms:
                    - my-term:
                        source_port: 1234
                        source_address:
                            - 1.2.3.4/32
                            - 5.6.7.8/32

        The ``pillar_key`` field would be specified as ``firewall``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``True``
        Merge the CLI variables with the pillar. Default: ``True``.

    revision_id
        Add a comment in the term config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the term configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    source_service
        A special service to choose from. This is a helper so the user is able to
        select a source just using the name, instead of specifying a source_port and protocol.

        As this module is available on Unix platforms only,
        it reads the IANA_ port assignment from ``/etc/services``.

        If the user requires additional shortcuts to be referenced, they can add entries under ``/etc/services``,
        which can be managed using the :mod:`file state <salt.states.file>`.

        .. _IANA: http://www.iana.org/assignments/port-numbers

    destination_service
        A special service to choose from. This is a helper so the user is able to
        select a source just using the name, instead of specifying a destination_port and protocol.
        Allows the same options as ``source_service``.

    term_fields
        Term attributes.
        To see what fields are supported, please consult the list of supported keywords_.
        Some platforms have few other optional_ keywords.

        .. _keywords: https://github.com/google/capirca/wiki/Policy-format#keywords
        .. _optional: https://github.com/google/capirca/wiki/Policy-format#optionally-supported-keywords

    .. note::
        The following fields are accepted:

        - action
        - address
        - address_exclude
        - comment
        - counter
        - expiration
        - destination_address
        - destination_address_exclude
        - destination_port
        - destination_prefix
        - forwarding_class
        - forwarding_class_except
        - logging
        - log_name
        - loss_priority
        - option
        - policer
        - port
        - precedence
        - principals
        - protocol
        - protocol_except
        - qos
        - pan_application
        - routing_instance
        - source_address
        - source_address_exclude
        - source_port
        - source_prefix
        - verbatim
        - packet_length
        - fragment_offset
        - hop_limit
        - icmp_type
        - ether_type
        - traffic_class_count
        - traffic_type
        - translated
        - dscp_set
        - dscp_match
        - dscp_except
        - next_ip
        - flexible_match_range
        - source_prefix_except
        - destination_prefix_except
        - vpn
        - source_tag
        - destination_tag
        - source_interface
        - destination_interface
        - flattened
        - flattened_addr
        - flattened_saddr
        - flattened_daddr
        - priority

    .. note::
        The following fields can be also a single value and a list of values:

        - action
        - address
        - address_exclude
        - comment
        - destination_address
        - destination_address_exclude
        - destination_port
        - destination_prefix
        - forwarding_class
        - forwarding_class_except
        - logging
        - option
        - port
        - precedence
        - principals
        - protocol
        - protocol_except
        - pan_application
        - source_address
        - source_address_exclude
        - source_port
        - source_prefix
        - verbatim
        - icmp_type
        - ether_type
        - traffic_type
        - dscp_match
        - dscp_except
        - flexible_match_range
        - source_prefix_except
        - destination_prefix_except
        - source_tag
        - destination_tag
        - source_service
        - destination_service

        Example: ``destination_address`` can be either defined as:

        .. code-block:: yaml

            destination_address: 172.17.17.1/24

        or as a list of destination IP addresses:

        .. code-block:: yaml

            destination_address:
                - 172.17.17.1/24
                - 172.17.19.1/24

        or a list of services to be matched:

        .. code-block:: yaml

            source_service:
                - ntp
                - snmp
                - ldap
                - bgpd

    .. note::
        The port fields ``source_port`` and ``destination_port`` can be used as above to select either
        a single value, either a list of values, but also they can select port ranges. Example:

        .. code-block:: yaml

            source_port:
                - [1000, 2000]
                - [3000, 4000]

        With the configuration above, the user is able to select the 1000-2000 and 3000-4000 source port ranges.

    CLI Example:

    .. code-block:: bash

        salt '*' capirca.get_term_config arista filter-name term-name source_address=1.2.3.4 destination_address=5.6.7.8 action=accept

    Output Example:

    .. code-block:: text

        ! $Date: 2017/03/22 $
        no ip access-list filter-name
        ip access-list filter-name
         remark term-name
         permit ip host 1.2.3.4 host 5.6.7.8
        exit
    """
    terms = []
    term = {term_name: {}}
    term[term_name].update(term_fields)
    term[term_name].update(
        {
            "source_service": _make_it_list({}, "source_service", source_service),
            "destination_service": _make_it_list(
                {}, "destination_service", destination_service
            ),
        }
    )
    terms.append(term)
    if not filter_options:
        filter_options = []
    return get_filter_config(
        platform,
        filter_name,
        filter_options=filter_options,
        terms=terms,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
        only_lower_merge=True,
        revision_id=revision_id,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
    )


def get_filter_config(
    platform,
    filter_name,
    filter_options=None,
    terms=None,
    prepend=True,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=True,
    only_lower_merge=False,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
):
    """
    Return the configuration of a policy filter.

    platform
        The name of the Capirca platform.

    filter_name
        The name of the policy filter.

    filter_options
        Additional filter options. These options are platform-specific.
        See the complete list of options_.

        .. _options: https://github.com/google/capirca/wiki/Policy-format#header-section

    terms
        List of terms for this policy filter.
        If not specified or empty, will try to load the configuration from the pillar,
        unless ``merge_pillar`` is set as ``False``.

    prepend: ``True``
        When ``merge_pillar`` is set as ``True``, the final list of terms generated by merging
        the terms from ``terms`` with those defined in the pillar (if any): new terms are prepended
        at the beginning, while existing ones will preserve the position. To add the new terms
        at the end of the list, set this argument to ``False``.

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``True``
        Merge the CLI variables with the pillar. Default: ``True``.

    only_lower_merge: ``False``
        Specify if it should merge only the terms fields. Otherwise it will try
        to merge also filters fields. Default: ``False``.

    revision_id
        Add a comment in the filter config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the filter configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    CLI Example:

    .. code-block:: bash

        salt '*' capirca.get_filter_config ciscoxr my-filter pillar_key=netacl

    Output Example:

    .. code-block:: text

        ! $Id:$
        ! $Date:$
        ! $Revision:$
        no ipv4 access-list my-filter
        ipv4 access-list my-filter
         remark $Id:$
         remark my-term
         deny ipv4 any eq 1234 any
         deny ipv4 any eq 1235 any
         remark my-other-term
         permit tcp any range 5678 5680 any
        exit

    The filter configuration has been loaded from the pillar, having the following structure:

    .. code-block:: yaml

        netacl:
          - my-filter:
              terms:
                - my-term:
                    source_port: [1234, 1235]
                    action: reject
                - my-other-term:
                    source_port:
                      - [5678, 5680]
                    protocol: tcp
                    action: accept
    """
    if not filter_options:
        filter_options = []
    if not terms:
        terms = []
    if merge_pillar and not only_lower_merge:
        acl_pillar_cfg = _get_pillar_cfg(
            pillar_key, saltenv=saltenv, pillarenv=pillarenv
        )
        filter_pillar_cfg = _lookup_element(acl_pillar_cfg, filter_name)
        filter_options = filter_options or filter_pillar_cfg.pop("options", None)
        if filter_pillar_cfg:
            # Only when it was able to find the filter in the ACL config
            pillar_terms = filter_pillar_cfg.get(
                "terms", []
            )  # No problem if empty in the pillar
            terms = _merge_list_of_dict(terms, pillar_terms, prepend=prepend)
            # merge the passed variable with the pillar data
            # any filter term not defined here, will be appended from the pillar
            # new terms won't be removed
    filters = []
    filters.append(
        {
            filter_name: {
                "options": _make_it_list({}, filter_name, filter_options),
                "terms": terms,
            }
        }
    )
    return get_policy_config(
        platform,
        filters=filters,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
        only_lower_merge=True,
        revision_id=revision_id,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
    )


def get_policy_config(
    platform,
    filters=None,
    prepend=True,
    pillar_key="acl",
    pillarenv=None,
    saltenv=None,
    merge_pillar=True,
    only_lower_merge=False,
    revision_id=None,
    revision_no=None,
    revision_date=True,
    revision_date_format="%Y/%m/%d",
):
    """
    Return the configuration of the whole policy.

    platform
        The name of the Capirca platform.

    filters
        List of filters for this policy.
        If not specified or empty, will try to load the configuration from the pillar,
        unless ``merge_pillar`` is set as ``False``.

    prepend: ``True``
        When ``merge_pillar`` is set as ``True``, the final list of filters generated by merging
        the filters from ``filters`` with those defined in the pillar (if any): new filters are prepended
        at the beginning, while existing ones will preserve the position. To add the new filters
        at the end of the list, set this argument to ``False``.

    pillar_key: ``acl``
        The key in the pillar containing the default attributes values. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.

    merge_pillar: ``True``
        Merge the CLI variables with the pillar. Default: ``True``.

    only_lower_merge: ``False``
        Specify if it should merge only the filters and terms fields. Otherwise it will try
        to merge everything at the policy level. Default: ``False``.

    revision_id
        Add a comment in the policy config having the description for the changes applied.

    revision_no
        The revision count.

    revision_date: ``True``
        Boolean flag: display the date when the policy configuration was generated. Default: ``True``.

    revision_date_format: ``%Y/%m/%d``
        The date format to be used when generating the perforce data. Default: ``%Y/%m/%d`` (<year>/<month>/<day>).

    CLI Example:

    .. code-block:: bash

        salt '*' capirca.get_policy_config juniper pillar_key=netacl

    Output Example:

    .. code-block:: text

        firewall {
            family inet {
                replace:
                /*
                ** $Id:$
                ** $Date:$
                ** $Revision:$
                **
                */
                filter my-filter {
                    term my-term {
                        from {
                            source-port [ 1234 1235 ];
                        }
                        then {
                            reject;
                        }
                    }
                    term my-other-term {
                        from {
                            protocol tcp;
                            source-port 5678-5680;
                        }
                        then accept;
                    }
                }
            }
        }
        firewall {
            family inet {
                replace:
                /*
                ** $Id:$
                ** $Date:$
                ** $Revision:$
                **
                */
                filter my-other-filter {
                    interface-specific;
                    term dummy-term {
                        from {
                            protocol [ tcp udp ];
                        }
                        then {
                            reject;
                        }
                    }
                }
            }
        }

    The policy configuration has been loaded from the pillar, having the following structure:

    .. code-block:: yaml

        netacl:
          - my-filter:
              options:
                - not-interface-specific
              terms:
                - my-term:
                    source_port: [1234, 1235]
                    action: reject
                - my-other-term:
                    source_port:
                      - [5678, 5680]
                    protocol: tcp
                    action: accept
          - my-other-filter:
              terms:
                - dummy-term:
                    protocol:
                      - tcp
                      - udp
                    action: reject
    """
    if not filters:
        filters = []
    if merge_pillar and not only_lower_merge:
        # the pillar key for the policy config is the `pillar_key` itself
        policy_pillar_cfg = _get_pillar_cfg(
            pillar_key, saltenv=saltenv, pillarenv=pillarenv
        )
        # now, let's merge everything witht the pillar data
        # again, this will not remove any extra filters/terms
        # but it will merge with the pillar data
        # if this behaviour is not wanted, the user can set `merge_pillar` as `False`
        filters = _merge_list_of_dict(filters, policy_pillar_cfg, prepend=prepend)
    policy_object = _get_policy_object(
        platform,
        filters=filters,
        pillar_key=pillar_key,
        pillarenv=pillarenv,
        saltenv=saltenv,
        merge_pillar=merge_pillar,
    )
    policy_text = str(policy_object)
    return _revision_tag(
        policy_text,
        revision_id=revision_id,
        revision_no=revision_no,
        revision_date=revision_date,
        revision_date_format=revision_date_format,
    )


def get_filter_pillar(filter_name, pillar_key="acl", pillarenv=None, saltenv=None):
    """
    Helper that can be used inside a state SLS,
    in order to get the filter configuration given its name.

    filter_name
        The name of the filter.

    pillar_key
        The root key of the whole policy config.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.
    """
    pillar_cfg = _get_pillar_cfg(pillar_key, pillarenv=pillarenv, saltenv=saltenv)
    return _lookup_element(pillar_cfg, filter_name)


def get_term_pillar(
    filter_name, term_name, pillar_key="acl", pillarenv=None, saltenv=None
):
    """
    Helper that can be used inside a state SLS,
    in order to get the term configuration given its name,
    under a certain filter uniquely identified by its name.

    filter_name
        The name of the filter.

    term_name
        The name of the term.

    pillar_key: ``acl``
        The root key of the whole policy config. Default: ``acl``.

    pillarenv
        Query the master to generate fresh pillar data on the fly,
        specifically from the requested pillar environment.

    saltenv
        Included only for compatibility with
        :conf_minion:`pillarenv_from_saltenv`, and is otherwise ignored.
    """
    filter_pillar_cfg = get_filter_pillar(
        filter_name, pillar_key=pillar_key, pillarenv=pillarenv, saltenv=saltenv
    )
    term_pillar_cfg = filter_pillar_cfg.get("terms", [])
    term_opts = _lookup_element(term_pillar_cfg, term_name)
    return term_opts



"""
:codeauthor: Herbert Buurman <herbert.buurman@ogd.nl>
"""


import datetime
import os
import textwrap

# Import Salt Module
import salt.modules.acme as acme
import salt.utils.dictupdate
import salt.utils.platform
from salt.exceptions import SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class AcmeTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.acme
    """

    def setup_loader_modules(self):
        return {acme: {}}

    def test_certs(self):
        """
        Test listing certs
        """
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "file.readdir": MagicMock(
                    return_value=[".", "..", "README", "test_expired", "test_valid"]
                )
            },
        ), patch(
            "os.path.isdir",
            side_effect=lambda path: path
            in [
                os.path.join(acme.LE_LIVE, "test_expired"),
                os.path.join(acme.LE_LIVE, "test_valid"),
            ],
        ):
            self.assertEqual(acme.certs(), ["test_expired", "test_valid"])

    def test_has(self):
        """
        Test checking if certificate (does not) exist.
        """
        with patch.dict(
            acme.__salt__, {"file.file_exists": MagicMock(return_value=True)}
        ):  # pylint: disable=no-member
            self.assertTrue(acme.has("test_expired"))
        with patch.dict(
            acme.__salt__, {"file.file_exists": MagicMock(return_value=False)}
        ):  # pylint: disable=no-member
            self.assertFalse(acme.has("test_invalid"))

    def test_needs_renewal(self):
        """
        Test if expired certs do indeed need renewal.
        """
        expired = (
            datetime.date.today()
            - datetime.timedelta(days=3)
            - datetime.date(1970, 1, 1)
        )
        valid = (
            datetime.date.today()
            + datetime.timedelta(days=3)
            - datetime.date(1970, 1, 1)
        )
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "tls.cert_info": MagicMock(
                    return_value={"not_after": expired.total_seconds()}
                )
            },
        ):
            self.assertTrue(acme.needs_renewal("test_expired"))
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "tls.cert_info": MagicMock(
                    return_value={"not_after": valid.total_seconds()}
                )
            },
        ):
            self.assertFalse(acme.needs_renewal("test_valid"))
            # Test with integer window parameter
            self.assertTrue(acme.needs_renewal("test_valid", window=5))
            # Test with string-like window parameter
            self.assertTrue(acme.needs_renewal("test_valid", window="5"))
            # Test with 'force' parameter
            self.assertTrue(acme.needs_renewal("test_valid", window="force"))
            # Test with 'true' parameter
            self.assertTrue(acme.needs_renewal("test_valid", window=True))
            # Test with invalid window parameter
            self.assertRaises(
                SaltInvocationError, acme.needs_renewal, "test_valid", window="foo"
            )

    def test_expires(self):
        """
        Test if expires function functions properly.
        """
        test_value = datetime.datetime.today() - datetime.timedelta(days=3)
        test_stamp = test_value - datetime.datetime(1970, 1, 1)
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "tls.cert_info": MagicMock(
                    return_value={"not_after": test_stamp.total_seconds()}
                )
            },
        ):
            self.assertEqual(
                acme.expires("test_expired"),
                datetime.datetime.fromtimestamp(test_stamp.total_seconds()).isoformat(),
            )

    def test_info(self):
        """
        Test certificate information retrieval.
        """
        certinfo_tls_result = {
            "not_after": 1559471377,
            "signature_algorithm": "sha256WithRSAEncryption",
            "extensions": {},
            "fingerprint": (
                "FB:A4:5F:71:D6:5D:6C:B6:1D:2C:FD:91:09:2C:1C:52:"
                "3C:EC:B6:4D:1A:95:65:37:04:D0:E2:5E:C7:64:0C:9C"
            ),
            "serial_number": 6461481982668892235,
            "issuer": {},
            "not_before": 1559557777,
            "subject": {},
        }
        certinfo_x509_result = {
            "Not After": "2019-06-02 10:29:37",
            "Subject Hash": "54:3B:6C:A4",
            "Serial Number": "59:AB:CB:A0:FB:90:E8:4B",
            "SHA1 Finger Print": (
                "F1:8D:F3:26:1B:D3:88:32:CD:B6:FA:3B:85:58:DA:C7:6F:62:BE:7E"
            ),
            "SHA-256 Finger Print": (
                "FB:A4:5F:71:D6:5D:6C:B6:1D:2C:FD:91:09:2C:1C:52:"
                "3C:EC:B6:4D:1A:95:65:37:04:D0:E2:5E:C7:64:0C:9C"
            ),
            "MD5 Finger Print": "95:B5:96:9B:42:A5:9E:20:78:FD:99:09:4B:21:1E:97",
            "Version": 3,
            "Key Size": 2048,
            "Public Key": (
                "-----BEGIN PUBLIC KEY-----\n"
                "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsVO2vwQPKU92PSBnuGid\n"
                "k8t6KWVE2jEBM10u7CgqQmD/JCnYflEHAo1nOsD7wxdhBrxhf5Qs+pEX1HOsh8VA\n"
                "HDTim0iE8nQVJ0Iuen2SrwaWMhwKmZTSJRYMgd46oCMi2RdlCvcgF2Hw6RTwF7FT\n"
                "hnksc4HBT91XddnP32N558tOT3YejafQNvClz5WcR+E0JzqGrV/+wfe3o+j/q5eK\n"
                "UowttWazeSMvuROtqj/fEk0rop4D14pgzZqWi30tjwhJNl6fSPFWBrLEHGNyDJ+O\n"
                "zfov0B2MRLJibH7GMkOCwsP2g1lVOReqcml+ju6zAKW8nHBTRg0iXB18Ifxef57Y\n"
                "AQIDAQAB\n"
                "-----END PUBLIC KEY-----\n"
            ),
            "Issuer": {},
            "Issuer Hash": "54:3B:6C:A4",
            "Not Before": "2019-06-03 10:29:37",
            "Subject": {},
        }

        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "file.file_exists": MagicMock(return_value=True),
                "tls.cert_info": MagicMock(return_value=certinfo_tls_result),
            },
        ):
            self.assertEqual(acme.info("test"), certinfo_tls_result)
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "file.file_exists": MagicMock(return_value=True),
                "x509.read_certificate": MagicMock(return_value=certinfo_x509_result),
            },
        ):
            self.assertEqual(acme.info("test"), certinfo_x509_result)
        with patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "file.file_exists": MagicMock(return_value=True),
                "cmd.run": MagicMock(return_value="foo"),
            },
        ):
            self.assertEqual(acme.info("test"), {"text": "foo"})

    def test_cert(self):
        """
        Test certificate retrieval/renewal
        """
        valid_timestamp = (
            datetime.datetime.now()
            + datetime.timedelta(days=30)
            - datetime.datetime(1970, 1, 1, 0, 0, 0, 0)
        ).total_seconds()
        expired_timestamp = (
            datetime.datetime.now()
            - datetime.timedelta(days=3)
            - datetime.datetime(1970, 1, 1, 0, 0, 0, 0)
        ).total_seconds()
        cmd_new_cert = {
            "stdout": textwrap.dedent(
                """
                IMPORTANT NOTES:
                 - Congratulations! Your certificate and chain have been saved at:
                   /etc/letsencrypt/live/test/fullchain.pem
                   Your key file has been saved at:
                   /etc/letsencrypt/live/test/privkey.pem
                   Your cert will expire on 2019-08-07. To obtain a new or tweaked
                   version of this certificate in the future, simply run certbot
                   again. To non-interactively renew *all* of your certificates, run
                   "certbot renew"
                 - If you like Certbot, please consider supporting our work by:

                   Donating to ISRG / Let's Encrypt:   https://letsencrypt.org/donate
                   Donating to EFF:                    https://eff.org/donate-le
                """
            ),
            "stderr": textwrap.dedent(
                """
                Saving debug log to /var/log/letsencrypt/letsencrypt.log
                Plugins selected: Authenticator standalone, Installer None
                Starting new HTTPS connection (1): acme-v02.api.letsencrypt.org
                Obtaining a new certificate
                Resetting dropped connection: acme-v02.api.letsencrypt.org
                """
            ),
            "retcode": 0,
        }
        result_new_cert = {
            "comment": "Certificate test obtained",
            "not_after": datetime.datetime.fromtimestamp(valid_timestamp).isoformat(),
            "changes": {"mode": "0640"},
            "result": True,
        }

        cmd_no_renew = {
            "stdout": textwrap.dedent(
                """
                - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                Certificate not yet due for renewal; no action taken.
                - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                """
            ),
            "stderr": textwrap.dedent(
                """Saving debug log to /var/log/letsencrypt/letsencrypt.log
                Plugins selected: Authenticator standalone, Installer None
                Starting new HTTPS connection (1): acme-v02.api.letsencrypt.org
                Cert not yet due for renewal
                Keeping the existing certificate
                """
            ),
            "retcode": 0,
        }
        if salt.utils.platform.is_freebsd():
            result_no_renew = {
                "comment": "Certificate "
                + os.path.join("/usr/local/etc/letsencrypt/live/test", "cert.pem")
                + " unchanged",
                "not_after": datetime.datetime.fromtimestamp(
                    valid_timestamp
                ).isoformat(),
                "changes": {},
                "result": True,
            }
        else:
            result_no_renew = {
                "comment": "Certificate "
                + os.path.join("/etc/letsencrypt/live/test", "cert.pem")
                + " unchanged",
                "not_after": datetime.datetime.fromtimestamp(
                    valid_timestamp
                ).isoformat(),
                "changes": {},
                "result": True,
            }
        result_renew = {
            "comment": "Certificate test renewed",
            "not_after": datetime.datetime.fromtimestamp(expired_timestamp).isoformat(),
            "changes": {},
            "result": True,
        }

        # Test fetching new certificate
        with patch("salt.modules.acme.LEA", "certbot"), patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "cmd.run_all": MagicMock(return_value=cmd_new_cert),
                "file.file_exists": MagicMock(return_value=False),
                "tls.cert_info": MagicMock(return_value={"not_after": valid_timestamp}),
                "file.check_perms": MagicMock(
                    side_effect=lambda a, x, b, c, d, follow_symlinks: (
                        salt.utils.dictupdate.set_dict_key_value(
                            x, "changes:mode", "0640"
                        ),
                        None,
                    )
                ),
            },
        ):
            self.assertEqual(acme.cert("test"), result_new_cert)
            self.assertEqual(
                acme.cert("testing.example.com", certname="test"), result_new_cert
            )
        # Test not renewing a valid certificate
        with patch("salt.modules.acme.LEA", "certbot"), patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "cmd.run_all": MagicMock(return_value=cmd_no_renew),
                "file.file_exists": MagicMock(return_value=True),
                "tls.cert_info": MagicMock(return_value={"not_after": valid_timestamp}),
                "file.check_perms": MagicMock(
                    side_effect=lambda a, x, b, c, d, follow_symlinks: (
                        salt.utils.dictupdate.set_dict_key_value(x, "result", True),
                        None,
                    )
                ),
            },
        ):
            self.assertEqual(acme.cert("test"), result_no_renew)
            self.assertEqual(
                acme.cert("testing.example.com", certname="test"), result_no_renew
            )
        # Test renewing an expired certificate
        with patch("salt.modules.acme.LEA", "certbot"), patch.dict(
            acme.__salt__,
            {  # pylint: disable=no-member
                "cmd.run_all": MagicMock(return_value=cmd_new_cert),
                "file.file_exists": MagicMock(return_value=True),
                "tls.cert_info": MagicMock(
                    return_value={"not_after": expired_timestamp}
                ),
                "file.check_perms": MagicMock(
                    side_effect=lambda a, x, b, c, d, follow_symlinks: (
                        salt.utils.dictupdate.set_dict_key_value(x, "result", True),
                        None,
                    )
                ),
            },
        ):
            self.assertEqual(acme.cert("test"), result_renew)
            self.assertEqual(
                acme.cert("testing.example.com", certname="test"), result_renew
            )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.states.ipmi as ipmi
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class IpmiTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.ipmi
    """

    def setup_loader_modules(self):
        return {ipmi: {}}

    # 'boot_device' function tests: 1

    def test_boot_device(self):
        """
        Test to request power state change.
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock = MagicMock(return_value=name)
        with patch.dict(
            ipmi.__salt__, {"ipmi.get_bootdev": mock, "ipmi.set_bootdev": mock}
        ):
            comt = "system already in this state"
            ret.update({"comment": comt})
            self.assertDictEqual(ipmi.boot_device(name), ret)

            with patch.dict(ipmi.__opts__, {"test": False}):
                comt = "changed boot device"
                ret.update(
                    {
                        "name": "default",
                        "comment": comt,
                        "result": True,
                        "changes": {"new": "default", "old": "salt"},
                    }
                )
                self.assertDictEqual(ipmi.boot_device(), ret)

            with patch.dict(ipmi.__opts__, {"test": True}):
                comt = "would change boot device"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(ipmi.boot_device(), ret)

    # 'power' function tests: 1

    def test_power(self):
        """
        Test to request power state change
        """
        ret = {"name": "power_on", "result": True, "comment": "", "changes": {}}

        mock = MagicMock(return_value="on")
        with patch.dict(
            ipmi.__salt__, {"ipmi.get_power": mock, "ipmi.set_power": mock}
        ):
            comt = "system already in this state"
            ret.update({"comment": comt})
            self.assertDictEqual(ipmi.power(), ret)

            with patch.dict(ipmi.__opts__, {"test": False}):
                comt = "changed system power"
                ret.update(
                    {
                        "name": "off",
                        "comment": comt,
                        "result": True,
                        "changes": {"new": "off", "old": "on"},
                    }
                )
                self.assertDictEqual(ipmi.power("off"), ret)

            with patch.dict(ipmi.__opts__, {"test": True}):
                comt = "would power: off system"
                ret.update({"comment": comt, "result": None})
                self.assertDictEqual(ipmi.power("off"), ret)

    # 'user_present' function tests: 1

    def test_user_present(self):
        """
        Test to ensure IPMI user and user privileges.
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock_ret = {
            "access": {
                "callback": False,
                "link_auth": True,
                "ipmi_msg": True,
                "privilege_level": "administrator",
            }
        }
        mock = MagicMock(return_value=mock_ret)
        mock_bool = MagicMock(side_effect=[True, False, False, False])
        with patch.dict(
            ipmi.__salt__,
            {
                "ipmi.get_user": mock,
                "ipmi.set_user_password": mock_bool,
                "ipmi.ensure_user": mock_bool,
            },
        ):
            comt = "user already present"
            ret.update({"comment": comt})
            self.assertDictEqual(ipmi.user_present(name, 5, "salt@123"), ret)

            with patch.dict(ipmi.__opts__, {"test": True}):
                comt = "would (re)create user"
                ret.update(
                    {
                        "comment": comt,
                        "result": None,
                        "changes": {"new": "salt", "old": mock_ret},
                    }
                )
                self.assertDictEqual(ipmi.user_present(name, 5, "pw@123"), ret)

            with patch.dict(ipmi.__opts__, {"test": False}):
                comt = "(re)created user"
                ret.update(
                    {
                        "comment": comt,
                        "result": True,
                        "changes": {"new": mock_ret, "old": mock_ret},
                    }
                )
                self.assertDictEqual(ipmi.user_present(name, 5, "pw@123"), ret)

    # 'user_absent' function tests: 1

    def test_user_absent(self):
        """
        Test to delete all user (uid) records having the matching name.
        """
        name = "salt"

        ret = {"name": name, "result": True, "comment": "", "changes": {}}

        mock = MagicMock(side_effect=[[], [5], [5]])
        mock_bool = MagicMock(return_value=True)
        with patch.dict(
            ipmi.__salt__, {"ipmi.get_name_uids": mock, "ipmi.delete_user": mock_bool}
        ):
            comt = "user already absent"
            ret.update({"comment": comt})
            self.assertDictEqual(ipmi.user_absent(name), ret)

            with patch.dict(ipmi.__opts__, {"test": True}):
                comt = "would delete user(s)"
                ret.update(
                    {"comment": comt, "result": None, "changes": {"delete": [5]}}
                )
                self.assertDictEqual(ipmi.user_absent(name), ret)

            with patch.dict(ipmi.__opts__, {"test": False}):
                comt = "user(s) removed"
                ret.update(
                    {
                        "comment": comt,
                        "result": False,
                        "changes": {"new": "None", "old": [5]},
                    }
                )
                self.assertDictEqual(ipmi.user_absent(name), ret)



"""
Read in the roster from a flat file using the renderer system
"""

import logging

import salt.config
import salt.loader
from salt.roster import get_roster_file
from salt.template import compile_template

log = logging.getLogger(__name__)


def targets(tgt, tgt_type="glob", **kwargs):
    """
    Return the targets from the flat yaml file, checks opts for location but
    defaults to /etc/salt/roster
    """
    template = get_roster_file(__opts__)

    rend = salt.loader.render(__opts__, {})
    raw = compile_template(
        template,
        rend,
        __opts__["renderer"],
        __opts__["renderer_blacklist"],
        __opts__["renderer_whitelist"],
        mask_value="passw*",
        **kwargs
    )
    conditioned_raw = {}
    for minion in raw:
        conditioned_raw[str(minion)] = salt.config.apply_sdb(raw[minion])
    return __utils__["roster_matcher.targets"](conditioned_raw, tgt, tgt_type, "ipv4")
